"""
Django App configuration for the xblocks-contrib Discussion XBlock.
"""

from django.apps import AppConfig


class DiscussionAppConfig(AppConfig):
    """
    Configuration for the Discussion XBlock Django application.

    Why are we registering this as a Django App?
    --------------------------------------------
    By default, Django's template engine (specifically the `app_directories` loader)
    only discovers `templates/` folders that belong to an app listed in `INSTALLED_APPS`.

    Since XBlocks are inherently isolated components, they are not typically added
    to the LMS/CMS global `INSTALLED_APPS`. As a result, any standard Django templates
    (.html files) inside this block would be invisible to the core Open edX platform
    when it attempts to route or render them natively.

    Instead of forcing repository consumers to manually modify `lms/envs/common.py`
    or use deployment-specific Tutor plugins to append to `TEMPLATES['DIRS']`, we
    leverage the Open edX Django App Plugin Framework (`plugin_app`). This hook
    dynamically injects this module into `INSTALLED_APPS` at runtime.

    This automatically exposes our `templates/` directory to the platform natively
    without manual intervention.

    References:

    - Open edX Plugin App Framework:
      https://docs.openedx.org/projects/edx-django-utils/en/latest/plugins/how_tos/how_to_create_a_plugin_app.html

    - Django App Directories Loader:
      https://docs.djangoproject.com/en/stable/ref/templates/api/#django.template.loaders.app_directories.Loader

    """

    name = "xblocks_contrib.discussion"
    label = "xblocks_contrib_discussion"
    verbose_name = "XBlocks Contrib - Discussion Templates"

    plugin_app = {
        "settings_config": {
            "lms.djangoapp": {
                "default": "",
            },
            "cms.djangoapp": {
                "default": "",
            },
        },
    }



""" Studio Metadata Mixin"""
from django.conf import settings
from xblock.core import XBlock
from xblock.fields import Dict, Float, Integer, List, Scope, String

from xblocks_contrib.video.exceptions import TranscriptNotFoundError
from xblocks_contrib.video.video_transcripts_utils import TranscriptExtensions, get_html5_ids
from xblocks_contrib.video.video_xfields import RelativeTime


class StudioMetadataMixin:
    """
    Mixin providing Studio metadata editing capabilities for XBlocks.
    """

    @property
    def non_editable_metadata_fields(self):
        """
        Return the list of fields that should not be editable in Studio.

        When overriding, be sure to append to the superclasses' list.
        """
        # We are not allowing editing of xblock tag and name fields at this time (for any component).
        return [XBlock.tags, XBlock.name]

    def _create_metadata_editor_info(self, field):
        """
        Creates the information needed by the metadata editor for a specific field.
        """

        def jsonify_value(field, json_choice):
            """
            Convert field value to JSON, if needed.
            """
            if isinstance(json_choice, dict):
                new_json_choice = dict(json_choice)  # make a copy so below doesn't change the original
                if "display_name" in json_choice:
                    new_json_choice["display_name"] = get_text(json_choice["display_name"])
                if "value" in json_choice:
                    new_json_choice["value"] = field.to_json(json_choice["value"])
            else:
                new_json_choice = field.to_json(json_choice)
            return new_json_choice

        def get_text(value):
            """Localize a text value that might be None."""
            if value is None:
                return None
            else:
                return self.runtime.service(self, "i18n").ugettext(value)

        # gets the 'default_value' and 'explicitly_set' attrs
        metadata_field_editor_info = self.runtime.get_field_provenance(self, field)
        metadata_field_editor_info["field_name"] = field.name
        metadata_field_editor_info["display_name"] = get_text(field.display_name)
        metadata_field_editor_info["help"] = get_text(field.help)
        metadata_field_editor_info["value"] = field.read_json(self)

        # We support the following editors:
        # 1. A select editor for fields with a list of possible values (includes Booleans).
        # 2. Number editors for integers and floats.
        # 3. A generic string editor for anything else (editing JSON representation of the value).
        editor_type = "Generic"
        values = field.values
        if "values_provider" in field.runtime_options:
            values = field.runtime_options["values_provider"](self)
        if isinstance(values, (tuple, list)) and len(values) > 0:
            editor_type = "Select"
            values = [jsonify_value(field, json_choice) for json_choice in values]
        elif isinstance(field, Integer):
            editor_type = "Integer"
        elif isinstance(field, Float):
            editor_type = "Float"
        elif isinstance(field, List):
            editor_type = "List"
        elif isinstance(field, Dict):
            editor_type = "Dict"
        elif isinstance(field, RelativeTime):
            editor_type = "RelativeTime"
        elif isinstance(field, String) and field.name == "license":
            editor_type = "License"
        metadata_field_editor_info["type"] = editor_type
        metadata_field_editor_info["options"] = [] if values is None else values

        return metadata_field_editor_info

    def _get_editable_metadata_fields(self):
        """
        Returns the metadata fields to be edited in Studio. These are fields with scope `Scope.settings`.

        Can be limited by extending `non_editable_metadata_fields`.
        """
        metadata_fields = {}

        # Only use the fields from this class, not mixins
        fields = getattr(self, "unmixed_class", self.__class__).fields

        for field in fields.values():
            if field in self.non_editable_metadata_fields:
                continue
            if field.scope not in (Scope.settings, Scope.content):
                continue

            metadata_fields[field.name] = self._create_metadata_editor_info(field)

        return metadata_fields

    @property
    def editable_metadata_fields(self):
        """
        Returns the metadata fields to be edited in Studio.
        """
        editable_fields = self._get_editable_metadata_fields()

        settings_service = self.runtime.service(self, 'settings')
        if settings_service:
            xb_settings = settings_service.get_settings_bucket(self)
            if not xb_settings.get("licensing_enabled", False) and "license" in editable_fields:
                del editable_fields["license"]

        # Default Timed Transcript a.k.a `sub` has been deprecated and end users shall
        # not be able to modify it.
        editable_fields.pop('sub')

        languages = [{'label': label, 'code': lang} for lang, label in settings.ALL_LANGUAGES]
        languages.sort(key=lambda lang_item: lang_item['label'])
        editable_fields['transcripts']['custom'] = True
        editable_fields['transcripts']['languages'] = languages
        editable_fields['transcripts']['type'] = 'VideoTranslations'

        # We need to send ajax requests to show transcript status
        # whenever edx_video_id changes on frontend. Thats why we
        # are changing type to `VideoID` so that a specific
        # Backbonjs view can handle it.
        editable_fields['edx_video_id']['type'] = 'VideoID'

        # `public_access` is a boolean field and by default backbonejs code render it as a dropdown with 2 options
        # but in our case we also need to show an input field with dropdown, the input field will show the url to
        # be shared with leaners. This is not possible with default rendering logic in backbonjs code, that is why
        # we are setting a new type and then do a custom rendering in backbonejs code to render the desired UI.
        editable_fields['public_access']['type'] = 'PublicAccess'
        editable_fields['public_access']['url'] = self.get_public_video_url()

        # construct transcripts info and also find if `en` subs exist
        transcripts_info = self.get_transcripts_info()
        possible_sub_ids = [self.sub, self.youtube_id_1_0] + get_html5_ids(self.html5_sources)
        video_config_service = self.runtime.service(self, 'video_config')
        if video_config_service:
            for sub_id in possible_sub_ids:
                try:
                    _, sub_id, _ = video_config_service.get_transcript(
                        self, lang='en', output_format=TranscriptExtensions.TXT
                    )
                    transcripts_info['transcripts'] = dict(transcripts_info['transcripts'], en=sub_id)
                    break
                except TranscriptNotFoundError:
                    continue

        editable_fields['transcripts']['value'] = transcripts_info['transcripts']
        editable_fields['transcripts']['urlRoot'] = self.runtime.handler_url(
            self,
            'studio_transcript',
            'translation'
        ).rstrip('/?')
        editable_fields['handout']['type'] = 'FileUploader'

        return editable_fields



#!/usr/bin/env python
"""
Commandline tool for doing operations on Problems
"""

import argparse
import logging
import sys
from io import BytesIO

from calc import UndefinedVariable
from django.template.loader import get_template

from xblocks_contrib.problem.capa.capa_problem import LoncapaProblem

logging.basicConfig(format="%(levelname)s %(message)s")
log = logging.getLogger("capa.checker")


class DemoSystem:
    """Render templates using Django's template engine."""

    def __init__(self):
        self.DEBUG = True  # pylint: disable=invalid-name

    def render_template(self, template_filename, dictionary):
        """
        Render the specified template with the given dictionary of context data.
        """
        return get_template(template_filename).render(dictionary)


def main():
    """Parse command-line arguments to test or display Loncapa problem files."""

    parser = argparse.ArgumentParser(description="Check Problem Files")
    parser.add_argument("command", choices=["test", "show"])  # Watch? Render? Open?
    parser.add_argument("files", nargs="+", type=argparse.FileType("r"))
    parser.add_argument("--seed", required=False, type=int)
    parser.add_argument(
        "--log-level",
        required=False,
        default="INFO",
        choices=["info", "debug", "warn", "error", "INFO", "DEBUG", "WARN", "ERROR"],
    )

    args = parser.parse_args()
    log.setLevel(args.log_level.upper())

    system = DemoSystem()

    for problem_file in args.files:
        log.info("Opening %s", problem_file.name)

        try:
            problem = LoncapaProblem(  # pylint: disable=unexpected-keyword-arg, no-value-for-parameter
                problem_file,
                "fakeid",
                seed=args.seed,
                system=system,
            )
        except Exception as ex:  # pylint: disable=broad-exception-caught
            log.error("Could not parse file %s", problem_file.name)
            log.exception(ex)
            continue

        if args.command == "test":
            command_test(problem)
        elif args.command == "show":
            command_show(problem)

        problem_file.close()

    # In case we want to do anything else here.


def command_show(problem):
    """Display the text for this problem"""
    print(problem.get_html())


def command_test(problem):
    """Run tests on a problem while capturing and logging stdout and stderr output."""

    # We're going to trap stdout/stderr from the problems (yes, some print)
    old_stdout, old_stderr = sys.stdout, sys.stderr
    try:
        sys.stdout = BytesIO()
        sys.stderr = BytesIO()

        check_that_suggested_answers_work(problem)
        check_that_blanks_fail(problem)

        log_captured_output(sys.stdout, f"captured stdout from {problem}")
        log_captured_output(sys.stderr, f"captured stderr from {problem}")
    except Exception as e:  # pylint: disable=broad-exception-caught
        log.exception(e)
    finally:
        sys.stdout, sys.stderr = old_stdout, old_stderr


def check_that_blanks_fail(problem):
    """Leaving it blank should never work. Neither should a space."""
    blank_answers = dict((answer_id, "") for answer_id in problem.get_question_answers())
    grading_results = problem.grade_answers(blank_answers)
    try:
        assert all(result == "incorrect" for result in grading_results.values())
    except AssertionError:
        log.error(
            "Blank accepted as correct answer in %s for %s",
            problem,
            [answer_id for answer_id, result in sorted(grading_results.items()) if result != "incorrect"],
        )


def check_that_suggested_answers_work(problem):
    """
    Split this up so that we're only used for formula/numeric answers.

    Examples of where this fails:

    * Displayed answers use units but acceptable ones do not.
      - L1e0.xml
      - Presents itself as UndefinedVariable (when it tries to pass to calc)

    * "a or d" is what's displayed, but only "a" or "d" is accepted, not the
      string "a or d".
      - L1-e00.xml
    """
    # These are actual answers we get from the responsetypes
    real_answers = problem.get_question_answers()

    # all_answers is real_answers + blanks for other answer_ids for which the
    # responsetypes can't provide us pre-canned answers (customresponse)
    all_answer_ids = problem.get_answer_ids()
    all_answers = dict((answer_id, real_answers.get(answer_id, "")) for answer_id in all_answer_ids)

    log.debug("Real answers: %s", real_answers)
    if real_answers:
        try:
            real_results = dict(
                (answer_id, result)
                for answer_id, result in problem.grade_answers(all_answers).items()
                if answer_id in real_answers
            )
            log.debug(real_results)
            assert all(result == "correct" for answer_id, result in real_results.items())
        except UndefinedVariable as uv_exc:
            log.error(
                'The variable "%s" specified in the solution isn\'t recognized (is it a units measure?).',
                uv_exc,
            )
        except AssertionError:
            log.error("The following generated answers were not accepted for %s:", problem)
            for question_id, result in sorted(real_results.items()):
                if result != "correct":
                    log.error("  %s = %s", question_id, real_answers[question_id])
        except Exception as ex:  # pylint: disable=broad-exception-caught
            log.error("Uncaught error in %s", problem)
            log.exception(ex)


def log_captured_output(output_stream, stream_name):
    """Log the contents of a captured output stream with header and footer markers."""

    output_stream.seek(0)
    output_text = output_stream.read()
    if output_text:
        log.info("##### Begin %s #####\n%s", stream_name, output_text)
        log.info("##### End %s #####", stream_name)


if __name__ == "__main__":
    sys.exit(main())



"""Utility functions for PDF XBlock."""

from django.conf import settings


def bool_from_str(str_value):
    """Convert string from submitted form to boolean."""
    return str_value.strip().lower() == 'true'


def is_all_download_disabled():
    """Check if all downloads are disabled or not."""
    return getattr(settings, 'PDFXBLOCK_DISABLE_ALL_DOWNLOAD', False)



"""Tools for helping with testing capa."""

import gettext
import io
import os
import os.path
from unittest.mock import MagicMock, Mock
from xml.sax import saxutils

import fs.osfs
from django.template.loader import get_template as django_get_template

from xblocks_contrib.problem.capa.capa_problem import LoncapaProblem, LoncapaSystem
from xblocks_contrib.problem.capa.inputtypes import Status

TEST_DIR = os.path.dirname(os.path.realpath(__file__))


def get_template(template_name):
    """
    Return template for a capa inputtype.
    """
    return django_get_template(template_name)


def capa_render_template(template, context):
    """
    Render template for a capa inputtype.
    """
    return get_template(template).render(context)


def tst_render_template(template, context):  # pylint: disable=unused-argument
    """
    A test version of render to template.  Renders to the repr of the context, completely ignoring
    the template name.  To make the output valid xml, quotes the content, and wraps it in a <div>
    """
    return f"<div>{saxutils.escape(repr(context))}</div>"


class StubXQueueService:
    """
    Stubs out the XQueueService for Capa problem tests.
    """

    def __init__(self):
        """Initialize the stubbed XQueueService instance."""
        self.interface = MagicMock()
        self.interface.send_to_queue.return_value = (0, "Success!")
        self.default_queuename = "testqueue"
        self.waittime = 10

    def construct_callback(self, dispatch="score_update"):
        """A callback url method to use in tests."""
        return dispatch


def mock_capa_system(render_template=None):
    """
    Construct a mock LoncapaSystem instance.
    """
    the_system = Mock(
        spec=LoncapaSystem,
        ajax_url="/dummy-ajax-url",
        anonymous_student_id="student",
        cache=None,
        can_execute_unsafe_code=lambda: False,
        get_python_lib_zip=lambda: None,
        DEBUG=True,
        i18n=gettext.NullTranslations(),
        render_template=render_template or tst_render_template,
        resources_fs=fs.osfs.OSFS(os.path.join(TEST_DIR, "test_files")),
        seed=0,
        STATIC_URL="/dummy-static/",
        STATUS_CLASS=Status,
        xqueue=StubXQueueService(),
    )
    return the_system


def mock_capa_block():
    """
    capa response types needs just two things from the capa_block: location and publish.
    """

    def mock_location_text(self):  # pylint: disable=unused-argument
        """
        Mock implementation of __unicode__ or __str__ for the block's location.
        """
        return "i4x://Foo/bar/mock/abc"

    capa_block = Mock()
    capa_block.usage_key.__str__ = mock_location_text
    # The following comes into existence by virtue of being called
    # capa_block.runtime.publish
    return capa_block


def new_loncapa_problem(xml, problem_id="1", capa_system=None, seed=723, use_capa_render_template=False):
    """Construct a `LoncapaProblem` suitable for unit tests."""
    render_template = capa_render_template if use_capa_render_template else None
    return LoncapaProblem(
        xml,
        id=problem_id,
        seed=seed,
        capa_system=capa_system or mock_capa_system(render_template),
        capa_block=mock_capa_block(),
    )


def load_fixture(relpath):
    """
    Return a `unicode` object representing the contents
    of the fixture file at the given path within a test_files directory
    in the same directory as the test file.
    """
    abspath = os.path.join(os.path.dirname(__file__), "test_files", relpath)
    with io.open(abspath, encoding="utf-8") as fixture_file:
        contents = fixture_file.read()
        return contents



"""  # lint-amnesty, pylint: disable=cyclic-import
XFields for video block.
"""


import datetime

from xblock.fields import Boolean, DateTime, Dict, Float, List, RelativeTime, Scope, String


# Make '_' a no-op so we can scrape strings. Using dummy function instead of
#  `django.utils.translation.gettext_noop` because Django cannot be imported in this file
def _(text):
    return text


class VideoFields:
    """Fields for `VideoBlock`."""
    data = String(default="", scope=Scope.content)

    display_name = String(
        help=_("The display name for this component."),
        display_name=_("Component Display Name"),
        default="Video",
        scope=Scope.settings
    )

    saved_video_position = RelativeTime(
        help=_("Current position in the video."),
        scope=Scope.user_state,
        default=datetime.timedelta(seconds=0)
    )
    # TODO: This should be moved to Scope.content, but this will
    # require data migration to support the old video block.
    youtube_id_1_0 = String(
        help=_("Optional, for older browsers: the YouTube ID for the normal speed video."),
        display_name=_("YouTube ID"),
        scope=Scope.settings,
        default="3_yD_cEKoCk"
    )
    youtube_id_0_75 = String(
        help=_("Optional, for older browsers: the YouTube ID for the .75x speed video."),
        display_name=_("YouTube ID for .75x speed"),
        scope=Scope.settings,
        default=""
    )
    youtube_id_1_25 = String(
        help=_("Optional, for older browsers: the YouTube ID for the 1.25x speed video."),
        display_name=_("YouTube ID for 1.25x speed"),
        scope=Scope.settings,
        default=""
    )
    youtube_id_1_5 = String(
        help=_("Optional, for older browsers: the YouTube ID for the 1.5x speed video."),
        display_name=_("YouTube ID for 1.5x speed"),
        scope=Scope.settings,
        default=""
    )
    start_time = RelativeTime(  # datetime.timedelta object
        help=_(
            "Time you want the video to start if you don't want the entire video to play. "
            "Not supported in the native mobile app: the full video file will play. "
            "Formatted as HH:MM:SS. The maximum value is 23:59:59."
        ),
        display_name=_("Video Start Time"),
        scope=Scope.settings,
        default=datetime.timedelta(seconds=0)
    )
    end_time = RelativeTime(  # datetime.timedelta object
        help=_(
            "Time you want the video to stop if you don't want the entire video to play. "
            "Not supported in the native mobile app: the full video file will play. "
            "Formatted as HH:MM:SS. The maximum value is 23:59:59."
        ),
        display_name=_("Video Stop Time"),
        scope=Scope.settings,
        default=datetime.timedelta(seconds=0)
    )
    # front-end code of video player checks logical validity of (start_time, end_time) pair.

    download_video = Boolean(
        help=_("Allow students to download versions of this video in different formats if they cannot use the edX video"
               " player or do not have access to YouTube. You must add at least one non-YouTube URL "
               "in the Video File URLs field."),
        display_name=_("Video Download Allowed"),
        scope=Scope.settings,
        default=False
    )
    html5_sources = List(
        help=_("The URL or URLs where you've posted non-YouTube versions of the video. Each URL must end in .mpeg,"
               " .mp4, .ogg, or .webm and cannot be a YouTube URL. (For browser compatibility, we strongly recommend"
               " .mp4 and .webm format.) Students will be able to view the first listed video that's compatible with"
               " the student's computer. To allow students to download these videos, "
               "set Video Download Allowed to True."),
        display_name=_("Video File URLs"),
        scope=Scope.settings,
    )
    track = String(
        help=_("By default, students can download an .srt or .txt transcript when you set Download Transcript "
               "Allowed to True. If you want to provide a downloadable transcript in a different format, we recommend "
               "that you upload a handout by using the Upload a Handout field. If this isn't possible, you can post a "
               "transcript file on the Files & Uploads page or on the Internet, and then add the URL for the "
               "transcript here. Students see a link to download that transcript below the video."),
        display_name=_("Downloadable Transcript URL"),
        scope=Scope.settings,
        default=''
    )
    download_track = Boolean(
        help=_("Allow students to download the timed transcript. A link to download the file appears below the video."
               " By default, the transcript is an .srt or .txt file. If you want to provide the transcript for "
               "download in a different format, upload a file by using the Upload Handout field."),
        display_name=_("Download Transcript Allowed"),
        scope=Scope.settings,
        default=False
    )
    # `sub` is deprecated field and should not be used in future. Now, transcripts are primarily handled in VAL and
    # backward compatibility for the video blocks already using this field has been ensured.
    sub = String(
        help=_("The default transcript for the video, from the Default Timed Transcript field on the Basic tab. "
               "This transcript should be in English. You don't have to change this setting."),
        display_name=_("Default Timed Transcript"),
        scope=Scope.settings,
        default=""
    )
    show_captions = Boolean(
        help=_("Specify whether the transcripts appear with the video by default."),
        display_name=_("Show Transcript"),
        scope=Scope.settings,
        default=True
    )
    # Data format: {'de': 'german_translation', 'uk': 'ukrainian_translation'}
    transcripts = Dict(
        help=_("Add transcripts in different languages."
               " Click below to specify a language and upload an .srt transcript file for that language."),
        display_name=_("Transcript Languages"),
        scope=Scope.settings,
        default={}
    )
    transcript_language = String(
        help=_("Preferred language for transcript."),
        display_name=_("Preferred language for transcript"),
        scope=Scope.preferences,
        default="en"
    )
    transcript_download_format = String(
        help=_("Transcript file format to download by user."),
        scope=Scope.preferences,
        values=[
            # Translators: This is a type of file used for captioning in the video player.
            {"display_name": _("SubRip (.srt) file"), "value": "srt"},
            {"display_name": _("Text (.txt) file"), "value": "txt"}
        ],
        default='srt',
    )
    speed = Float(
        help=_("The last speed that the user specified for the video."),
        scope=Scope.user_state
    )
    global_speed = Float(
        help=_("The default speed for the video."),
        scope=Scope.preferences,
        default=1.0
    )
    auto_advance = Boolean(
        help=_("Specify whether to advance automatically to the next unit when the video ends."),
        scope=Scope.preferences,
        # The default is True because this field only has an effect when auto-advance controls are enabled
        # (globally enabled through feature flag and locally enabled through course setting); in that case
        # it's good to start auto-advancing and let the student disable it, instead of the other way around
        # (requiring the user to enable it). When auto-advance controls are hidden, this field won't be used.
        default=True,
    )
    youtube_is_available = Boolean(
        help=_("Specify whether YouTube is available for the user."),
        scope=Scope.user_info,
        default=True
    )
    handout = String(
        help=_("Upload a handout to accompany this video. Students can download the handout by "
               "clicking Download Handout under the video."),
        display_name=_("Upload Handout"),
        scope=Scope.settings,
    )
    only_on_web = Boolean(
        help=_(
            "Specify whether access to this video is limited to browsers only, or if it can be "
            "accessed from other applications including mobile apps."
        ),
        display_name=_("Video Available on Web Only"),
        scope=Scope.settings,
        default=False
    )
    edx_video_id = String(
        help=_("If you were assigned a Video ID by edX for the video to play in this component, enter the ID here."
               " In this case, do not enter values in the Default Video URL, the Video File URLs, "
               "and the YouTube ID fields. If you were not assigned a Video ID,"
               " enter values in those other fields and ignore this field."),
        display_name=_("Video ID"),
        scope=Scope.settings,
        default="",
    )
    bumper_last_view_date = DateTime(
        display_name=_("Date of the last view of the bumper"),
        scope=Scope.preferences,
    )
    bumper_do_not_show_again = Boolean(
        display_name=_("Do not show bumper again"),
        scope=Scope.preferences,
        default=False,
    )
    public_access = Boolean(
        help=_("Specify whether the video can be accessed publicly by learners."),
        display_name=_("Public Access"),
        scope=Scope.settings,
        default=False
    )
    # thumbnail is need as a field for the new video editor. The field is hidden in
    # the legacy modal.
    thumbnail = String(
        help=_("Add a specific thumbnail for learners to see before playing the video."),
        display_name=_("Thumbnail"),
        scope=Scope.settings,
        default="",
    )



"""
Discussion XBlock
"""

import logging
import urllib

import markupsafe
from django.conf import settings
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import get_language_bidi
from web_fragments.fragment import Fragment
from xblock.completable import XBlockCompletionMode
from xblock.core import XBlock
from xblock.fields import UNIQUE_ID, Scope, String
from xblock.utils.resources import ResourceLoader
from xblock.utils.studio_editable import StudioEditableXBlockMixin

from xblocks_contrib.common.xml_utils import LegacyXmlMixin

log = logging.getLogger(__name__)
loader = ResourceLoader(__name__)
Text = markupsafe.escape                        # pylint: disable=invalid-name


def _(text):
    """
    A noop underscore function that marks strings for extraction.
    """
    return text


def HTML(html):                                 # pylint: disable=invalid-name
    """
    Mark a string as already HTML, so that it won't be escaped before output.

    Use this function when formatting HTML into other strings.  It must be
    used in conjunction with ``Text()``, and both ``HTML()`` and ``Text()``
    must be closed before any calls to ``format()``::

        <%page expression_filter="h"/>
        <%!
        from django.utils.translation import gettext as _

        from openedx.core.djangolib.markup import HTML, Text
        %>
        ${Text(_("Write & send {start}email{end}")).format(
            start=HTML("<a href='mailto:{}'>").format(user.email),
            end=HTML("</a>"),
        )}

    """
    return markupsafe.Markup(html)


@XBlock.needs("i18n")
@XBlock.needs("user")
@XBlock.wants('discussion_config_service')
# pylint: disable=abstract-method
class DiscussionXBlock(XBlock, StudioEditableXBlockMixin, LegacyXmlMixin):
    """
    Provides a discussion forum that is inline with other content in the courseware.
    """
    is_extracted = True
    completion_mode = XBlockCompletionMode.EXCLUDED

    discussion_id = String(scope=Scope.settings, default=UNIQUE_ID)
    display_name = String(
        display_name=_("Display Name"),
        help=_("The display name for this component."),
        default="Discussion",
        scope=Scope.settings
    )
    discussion_category = String(
        display_name=_("Category"),
        default=_("Week 1"),
        help=_(
            "A category name for the discussion. "
            "This name appears in the left pane of the discussion forum for the course."
        ),
        scope=Scope.settings
    )
    discussion_target = String(
        display_name=_("Subcategory"),
        default="Topic-Level Student-Visible Label",
        help=_(
            "A subcategory name for the discussion. "
            "This name appears in the left pane of the discussion forum for the course."
        ),
        scope=Scope.settings
    )
    sort_key = String(scope=Scope.settings)

    editable_fields = ["display_name", "discussion_category", "discussion_target"]

    has_author_view = True  # Tells Studio to use author_view

    @property
    def discussion_config(self):
        """
        Returns discussion service.
        """
        return self.runtime.service(self, 'discussion_config_service')

    @property
    def is_visible(self):
        """
        Discussion Xblock does not support new OPEN_EDX provider
        """
        return self.discussion_config.is_discussion_visible(self.context_key)

    @property
    def django_user(self):
        """
        Returns django user associated with user currently interacting
        with the XBlock.
        """
        user_service = self.runtime.service(self, 'user')
        if not user_service:
            return None
        return user_service._django_user  # pylint: disable=protected-access

    def add_resource_urls(self, fragment):
        """
        Adds URLs for JS and CSS resources that this XBlock depends on to `fragment`.
        """

        css_file_path = (
            '/css/inline-discussion-rtl.css'
            if get_language_bidi()
            else '/css/inline-discussion.css'
        )

        # Determine how static assets should be served based on Django settings.
        # Open edX requires different asset paths for production vs. local development.
        pipeline = getattr(settings, "PIPELINE", {})
        use_pipeline = pipeline.get("PIPELINE_ENABLED", True) or not getattr(settings, "REQUIRE_DEBUG", False)

        # When the Django pipeline is active (production), XBlock assets are namespaced
        # using the package scope (e.g., "discussion/public").
        # When inactive (local dev fallback), they are served directly from "public".
        # https://github.com/openedx/openedx-platform/blob/master/openedx/core/lib/xblock_utils/__init__.py#L417
        base_path = "discussion/public" if use_pipeline else "public"

        fragment.add_css_url(self.runtime.local_resource_url(self, f"{base_path}{css_file_path}"))
        fragment.add_javascript_url(self.runtime.local_resource_url(self, f"{base_path}/js/discussion_bundle.js"))

    def has_permission(self, permission):
        """
        Encapsulates lms specific functionality, as `has_permission` is not
        importable outside of lms context, namely in tests.

        :param user:
        :param str permission: Permission
        :rtype: bool
        """
        if self.discussion_config:
            return self.discussion_config.has_permission(self.django_user, permission, self.context_key)
        else:
            return False

    def student_view(self, context=None):
        """
        Renders student view for LMS.
        """

        fragment = Fragment()

        if not self.is_visible:
            return fragment

        self.add_resource_urls(fragment)
        login_msg = ''

        if not self.django_user.is_authenticated:
            qs = urllib.parse.urlencode({
                'course_id': self.context_key,
                'enrollment_action': 'enroll',
                'email_opt_in': False,
            })
            login_msg = Text(_("You are not signed in. To view the discussion content, {sign_in_link} or "
                               "{register_link}, and enroll in this course.")).format(
                sign_in_link=HTML('<a href="{url}">{sign_in_label}</a>').format(
                    sign_in_label=_('sign in'),
                    url='{}?{}'.format(reverse('signin_user'), qs),
                ),
                register_link=HTML('<a href="/{url}">{register_label}</a>').format(
                    register_label=_('register'),
                    url='{}?{}'.format(reverse('register_user'), qs),
                ),
            )

        if self.discussion_config.is_discussion_enabled:
            context = {
                'discussion_id': self.discussion_id,
                'display_name': self.display_name if self.display_name else _("Discussion"),
                'user': self.django_user,
                'course_id': self.context_key,
                'discussion_category': self.discussion_category,
                'discussion_target': self.discussion_target,
                'can_create_thread': self.has_permission("create_thread"),
                'can_create_comment': self.has_permission("create_comment"),
                'can_create_subcomment': self.has_permission("create_sub_comment"),
                'login_msg': login_msg,
                'PLATFORM_NAME': settings.PLATFORM_NAME,
                'enable_discussion_home_panel': settings.FEATURES.get("ENABLE_DISCUSSION_HOME_PANEL", False),
            }
            fragment.add_content(
                render_to_string('discussion_templates/_discussion_inline.html', context)
            )

        fragment.initialize_js('DiscussionInlineBlock')

        return fragment

    def author_view(self, context=None):
        """
        Renders author view for Studio.
        """
        fragment = Fragment()
        context = {
            'discussion_id': self.discussion_id,
            'is_visible': self.is_visible,
        }
        fragment.add_content(
            loader.render_django_template('templates/_discussion_inline_studio.html', context)
        )
        return fragment

    def student_view_data(self):
        """
        Returns a JSON representation of the student_view of this XBlock.
        """
        return {'topic_id': self.discussion_id}

    @classmethod
    def parse_xml(cls, node, runtime, keys):
        """
        Parses OLX into XBlock.

        This method is overridden here to allow parsing legacy OLX, coming from discussion XModule.
        XBlock stores all the associated data, fields and children in a XML element inlined into vertical XML file
        XModule stored only minimal data on the element included into vertical XML and used a dedicated "discussion"
        folder in OLX to store fields and children. Also, some info was put into "policy.json" file.

        If no external data sources are found (file in "discussion" folder), it is exactly equivalent to base method
        XBlock.parse_xml. Otherwise this method parses file in "discussion" folder (known as definition_xml), applies
        policy.json and updates fields accordingly.
        """
        block = super().parse_xml(node, runtime, keys)

        cls._apply_metadata_and_policy(block, node, runtime)

        return block

    @classmethod
    def _apply_metadata_and_policy(cls, block, node, runtime):
        """
        Attempt to load definition XML from "discussion" folder in OLX, than parse it and update block fields
        """
        if node.get('url_name') is None:
            return  # Newer/XBlock XML format - no need to load an additional file.
        try:
            definition_xml, _ = cls.load_definition_xml(node, runtime, block.scope_ids.def_id)
        except Exception as err:  # pylint: disable=broad-except
            log.info(
                "Exception %s when trying to load definition xml for block %s - assuming XBlock export format",
                err,
                block
            )
            return

        metadata = cls.load_metadata(definition_xml)
        cls.apply_policy(metadata, runtime.get_policy(block.usage_key))

        for field_name, value in metadata.items():
            if field_name in block.fields:
                setattr(block, field_name, value)



"""
Utilities for use in markup.
"""

import markupsafe

# Text() can be used to declare a string as plain text, as HTML() is used
# for HTML.  It simply wraps markupsafe's escape, which will HTML-escape if
# it isn't already escaped.
Text = markupsafe.escape  # pylint: disable=invalid-name


def HTML(html):  # pylint: disable=invalid-name
    """
    Mark a string as already HTML, so that it won't be escaped before output.

    Use this function when formatting HTML into other strings.  It must be
    used in conjunction with ``Text()``, and both ``HTML()`` and ``Text()``
    must be closed before any calls to ``format()``::

        <%page expression_filter="h"/>
        <%!
        from django.utils.translation import gettext as _

        from xblocks_contrib.problem.markup import HTML, Text
        %>
        ${Text(_("Write & send {start}email{end}")).format(
            start=HTML("<a href='mailto:{}'>").format(user.email),
            end=HTML("</a>"),
        )}

    """
    return markupsafe.Markup(html)



"""
Extension of XBlock Validation class to include information for presentation in Studio.
"""
from xblock.validation import Validation, ValidationMessage


class StudioValidationMessage(ValidationMessage):
    """
    A message containing validation information about an xblock, extended to provide Studio-specific fields.
    """

    # A special message type indicating that the xblock is not yet configured. This message may be rendered
    # in a different way within Studio.
    NOT_CONFIGURED = "not-configured"

    TYPES = [ValidationMessage.WARNING, ValidationMessage.ERROR, NOT_CONFIGURED]

    def __init__(  # pylint: disable=too-many-positional-arguments
        self, message_type, message_text, action_label=None, action_class=None, action_runtime_event=None
    ):
        """
        Create a new message.

        Args:
            message_type (str): The type associated with this message. Most be `WARNING` or `ERROR`.
            message_text (unicode): The textual message.
            action_label (unicode): Text to show on a "fix-up" action (optional). If present, either `action_class`
                or `action_runtime_event` should be specified.
            action_class (str): A class to link to the "fix-up" action (optional). A click handler must be added
                for this class, unless it is "edit-button", "duplicate-button", or "delete-button" (which are all
                handled in general for xblock instances.
            action_runtime_event (str): An event name to be triggered on the xblock client-side runtime when
                the "fix-up" action is clicked (optional).
        """
        super().__init__(message_type, message_text)
        if action_label is not None:
            if not isinstance(action_label, str):
                raise TypeError("Action label must be unicode.")
            self.action_label = action_label
        if action_class is not None:
            if not isinstance(action_class, str):
                raise TypeError("Action class must be a string.")
            self.action_class = action_class
        if action_runtime_event is not None:
            if not isinstance(action_runtime_event, str):
                raise TypeError("Action runtime event must be a string.")
            self.action_runtime_event = action_runtime_event

    def to_json(self):
        """
        Convert to a json-serializable representation.

        Returns:
            dict: A dict representation that is json-serializable.
        """
        serialized = super().to_json()
        if hasattr(self, "action_label"):
            serialized["action_label"] = self.action_label
        if hasattr(self, "action_class"):
            serialized["action_class"] = self.action_class
        if hasattr(self, "action_runtime_event"):
            serialized["action_runtime_event"] = self.action_runtime_event
        return serialized


class StudioValidation(Validation):
    """
    Extends `Validation` to add Studio-specific summary message.
    """

    @classmethod
    def copy(cls, validation):
        """
        Copies the `Validation` object to a `StudioValidation` object. This is a shallow copy.

        Args:
            validation (Validation): A `Validation` object to be converted to a `StudioValidation` instance.

        Returns:
            StudioValidation: A `StudioValidation` instance populated with the messages from supplied
                `Validation` object
        """
        if not isinstance(validation, Validation):
            raise TypeError("Copy must be called with a Validation instance")
        studio_validation = cls(validation.xblock_id)
        studio_validation.messages = validation.messages
        return studio_validation

    def __init__(self, xblock_id):
        """
        Create a `StudioValidation` instance.

        Args:
            xblock_id (object): An identification object that must support conversion to unicode.
        """
        super().__init__(xblock_id)
        self.summary = None

    def set_summary(self, message):
        """
        Sets a summary message on this instance. The summary is optional.

        Args:
            message (ValidationMessage): A validation message to set as this instance's summary.
        """
        if not isinstance(message, ValidationMessage):
            raise TypeError("Argument must of type ValidationMessage")
        self.summary = message

    @property
    def empty(self):
        """
        Is this object empty (contains no messages and no summary)?

        Returns:
            bool: True iff this instance has no validation issues and therefore has no messages or summary.
        """
        return super().empty and not self.summary

    def to_json(self):
        """
        Convert to a json-serializable representation.

        Returns:
            dict: A dict representation that is json-serializable.
        """
        serialized = super().to_json()
        if self.summary:
            serialized["summary"] = self.summary.to_json()
        return serialized



"""Utility for converting XML nodes to their inner string representation."""

from lxml import etree


def stringify_children(node):
    """
    Return all contents of an xml tree, without the outside tags.
    e.g. if node is parse of

        "<html a="b" foo="bar">Hi <div>there <span>Bruce</span><b>!</b></div><html>"

    should return

        "Hi <div>there <span>Bruce</span><b>!</b></div>"

    fixed from
    http://stackoverflow.com/questions/4624062/get-all-text-inside-a-tag-in-lxml
    """
    # Useful things to know:

    # node.tostring() -- generates xml for the node, including start
    #                 and end tags.  We'll use this for the children.
    # node.text -- the text after the end of a start tag to the start
    #                 of the first child
    # node.tail -- the text after the end this tag to the start of the
    #                 next element.
    parts = [node.text]
    for c in node.getchildren():
        parts.append(etree.tostring(c, with_tail=True, encoding="unicode"))

    # filter removes possible Nones in texts and tails
    return "".join([part for part in parts if part])



""" Cache Utils for Video Block """
import itertools

import wrapt
from django.utils.encoding import force_str
from edx_django_utils.cache import RequestCache


def request_cached(namespace=None, arg_map_function=None, request_cache_getter=None):
    """
    A function decorator that automatically handles caching its return value for
    the duration of the request. It returns the cached value for subsequent
    calls to the same function, with the same parameters, within a given request.

    Notes:
        - We convert arguments and keyword arguments to their string form to build the cache key. So if you have
          args/kwargs that can't be converted to strings, you're gonna have a bad time (don't do it).
        - Cache key cardinality depends on the args/kwargs. So if you're caching a function that takes five arguments,
          you might have deceptively low cache efficiency.  Prefer functions with fewer arguments.
        - WATCH OUT: Don't use this decorator for instance methods that take in a "self" argument that changes each
          time the method is called. This will result in constant cache misses and not provide the performance benefit
          you are looking for. Rather, change your instance method to a class method.
        - Benchmark, benchmark, benchmark! If you never measure, how will you know you've improved? or regressed?

    Arguments:
        namespace (string): An optional namespace to use for the cache. By default, we use the default request cache,
            not a namespaced request cache. Since the code automatically creates a unique cache key with the module and
            function's name, storing the cached value in the default cache, you won't usually need to specify a
            namespace value.
            But you can specify a namespace value here if you need to use your own namespaced cache - for example,
            if you want to clear out your own cache by calling RequestCache(namespace=NAMESPACE).clear().
            NOTE: This argument is ignored if you supply a ``request_cache_getter``.
        arg_map_function (function: arg->string): Function to use for mapping the wrapped function's arguments to
            strings to use in the cache key. If not provided, defaults to force_text, which converts the given
            argument to a string.
        request_cache_getter (function: args, kwargs->RequestCache): Function that returns the RequestCache to use.
            If not provided, defaults to edx_django_utils.cache.RequestCache.  If ``request_cache_getter`` returns None,
            the function's return values are not cached.

    Returns:
        func: a wrapper function which will call the wrapped function, passing in the same args/kwargs,
              cache the value it returns, and return that cached value for subsequent calls with the
              same args/kwargs within a single request.
    """
    @wrapt.decorator
    def decorator(wrapped, instance, args, kwargs):
        """
        Arguments:
            args, kwargs: values passed into the wrapped function
        """
        # Check to see if we have a result in cache.  If not, invoke our wrapped
        # function.  Cache and return the result to the caller.
        if request_cache_getter:
            request_cache = request_cache_getter(args if instance is None else (instance,) + args, kwargs)
        else:
            request_cache = RequestCache(namespace)

        if request_cache:
            cache_key = _func_call_cache_key(wrapped, arg_map_function, *args, **kwargs)
            cached_response = request_cache.get_cached_response(cache_key)
            if cached_response.is_found:
                return cached_response.value

        result = wrapped(*args, **kwargs)

        if request_cache:
            request_cache.set(cache_key, result)

        return result

    return decorator


def _func_call_cache_key(func, arg_map_function, *args, **kwargs):
    """
    Returns a cache key based on the function's module,
    the function's name, a stringified list of arguments
    and a stringified list of keyword arguments.
    """
    arg_map_function = arg_map_function or force_str

    converted_args = list(map(arg_map_function, args))
    converted_kwargs = list(map(arg_map_function, _sorted_kwargs_list(kwargs)))

    cache_keys = [func.__module__, func.__name__] + converted_args + converted_kwargs
    return '.'.join(cache_keys)


def _sorted_kwargs_list(kwargs):
    """
    Returns a unique and deterministic ordered list from the given kwargs.
    """
    sorted_kwargs = sorted(kwargs.items())
    sorted_kwargs_list = list(itertools.chain(*sorted_kwargs))
    return sorted_kwargs_list



"""
This module provides an interface for submitting student responses
to an external grading system through XQueue.
"""

import json
import logging

from xblocks_contrib.problem.capa.errors import (
    GetSubmissionParamsError,
    JSONParsingError,
    MissingKeyError,
    RuntimeErrorSubmission,
    TypeErrorSubmission,
    ValidationError,
)

log = logging.getLogger(__name__)


class XQueueInterfaceSubmission:
    """Interface to the external grading system."""

    def __init__(self, block):
        self.block = block

    def _parse_json(self, data, name):
        """
        Helper function to safely parse data that may or may not be a JSON string.
        This is necessary because some callers may already provide parsed Python dicts
        (e.g., during internal calls or test cases), while other sources may send raw JSON strings.
        This helper ensures consistent behavior regardless of input format.
        Args:
            data: The input to parse, either a JSON string or a Python dict.
            name: Name of the field (used for error reporting).
        Returns:
            Parsed Python object or original data if already parsed.
        Raises:
            JSONParsingError: If `data` is a string and cannot be parsed as JSON.
        """
        try:
            return json.loads(data) if isinstance(data, str) else data
        except json.JSONDecodeError as e:
            raise JSONParsingError(name, str(e)) from e

    def get_submission_params(self, header, payload):
        """
        Extracts student submission data from the given header and payload.
        """
        header = self._parse_json(header, "header")
        payload = self._parse_json(payload, "payload")

        queue_name = header.get("queue_name", "default")

        if not self.block:
            raise GetSubmissionParamsError()

        course_id = str(self.block.context_key)
        item_type = self.block.scope_ids.block_type
        points_possible = self.block.max_score()

        item_id = str(self.block.usage_key)

        try:
            grader_payload = self._parse_json(payload["grader_payload"], "grader_payload")
            grader_file_name = grader_payload.get("grader", "")
        except KeyError as e:
            raise MissingKeyError("grader_payload") from e

        student_info = self._parse_json(payload["student_info"], "student_info")
        student_id = student_info.get("anonymous_student_id")

        if not student_id:
            raise ValidationError("The field 'anonymous_student_id' is missing from student_info.")

        student_answer = payload.get("student_response")
        if student_answer is None:
            raise ValidationError("The field 'student_response' does not exist.")

        student_dict = {"item_id": item_id, "item_type": item_type, "course_id": course_id, "student_id": student_id}

        return student_dict, student_answer, queue_name, grader_file_name, points_possible

    def send_to_submission(self, header, body, queue_key, files_to_upload=None):
        """
        Submits the extracted student data to the edx-submissions system.
        """
        try:
            from submissions.api import create_external_grader_detail  # pylint: disable=import-outside-toplevel

            student_item, answer, queue_name, grader_file_name, points_possible = self.get_submission_params(
                header, body
            )
            return create_external_grader_detail(
                student_item,
                answer,
                queue_name=queue_name,
                queue_key=queue_key,
                grader_file_name=grader_file_name,
                points_possible=points_possible,
                files=files_to_upload,
            )
        except (JSONParsingError, MissingKeyError, ValidationError) as e:
            log.error("%s", e)
            return {"error": str(e)}
        except TypeError as e:
            log.error("%s", e)
            raise TypeErrorSubmission(str(e)) from e
        except RuntimeError as e:
            log.error("%s", e)
            raise RuntimeErrorSubmission(str(e)) from e



"""
License mixin for XBlocks
"""

from xblock.core import XBlockMixin
from xblock.fields import Scope, String


# Make '_' a no-op so we can scrape strings. Using dummy function instead of
#  `django.utils.translation.gettext_noop` because Django cannot be imported in this file
def _(text):
    return text


class LicenseMixin(XBlockMixin):
    """
    Mixin that allows an author to indicate a license on the contents of an
    XBlock. For example, a video could be marked as Creative Commons SA-BY
    licensed. You can even indicate the license on an entire course.

    If this mixin is not applied to an XBlock, or if the license field is
    blank, then the content is subject to whatever legal licensing terms that
    apply to content by default. For example, in the United States, that content
    is exclusively owned by the creator of the content by default. Other
    countries may have similar laws.
    """
    license = String(
        display_name=_("License"),
        help=_("A license defines how the contents of this block can be shared and reused."),
        default=None,
        scope=Scope.content,
    )

    @classmethod
    def parse_license_from_xml(cls, definition, node):
        """
        When importing an XBlock from XML, this method will parse the license
        information out of the XML and attach it to the block.
        It is defined here so that classes that use this mixin can simply refer
        to this method, rather than reimplementing it in their XML import
        functions.
        """
        license = node.get('license', default=None)  # pylint: disable=redefined-builtin
        if license:
            definition['license'] = license
        return definition

    def add_license_to_xml(self, node, default=None):
        """
        When generating XML from an XBlock, this method will add the XBlock's
        license to the XML representation before it is serialized.
        It is defined here so that classes that use this mixin can simply refer
        to this method, rather than reimplementing it in their XML export
        functions.
        """
        if getattr(self, "license", default):
            node.set('license', self.license)



"""
Helper methods related to safe exec.
"""

import json
import logging
from importlib import import_module

import requests
from codejail.safe_exec import SafeExecException, json_safe
from django.conf import settings
from django.utils.translation import gettext as _
from edx_toggles.toggles import SettingToggle
from requests.exceptions import HTTPError, RequestException
from simplejson import JSONDecodeError

from .exceptions import CodejailServiceParseError, CodejailServiceStatusError, CodejailServiceUnavailable

log = logging.getLogger(__name__)

# .. toggle_name: ENABLE_CODEJAIL_REST_SERVICE
# .. toggle_implementation: SettingToggle
# .. toggle_default: False
# .. toggle_description: Set this to True if you want to run Codejail code using
#   a separate VM or container and communicate with edx-platform using REST API.
# .. toggle_use_cases: open_edx
# .. toggle_creation_date: 2021-08-19
ENABLE_CODEJAIL_REST_SERVICE = SettingToggle("ENABLE_CODEJAIL_REST_SERVICE", default=False, module_name=__name__)

# .. toggle_name: ENABLE_CODEJAIL_DARKLAUNCH
# .. toggle_implementation: SettingToggle
# .. toggle_default: False
# .. toggle_description: Turn on to send requests to both the codejail service and the installed codejail library for
#   testing and evaluation purposes. The results from the installed codejail library will be the ones used.
# .. toggle_warning: This toggle will only behave as expected when ENABLE_CODEJAIL_REST_SERVICE is not enabled and when
#   CODE_JAIL_REST_SERVICE_REMOTE_EXEC, CODE_JAIL_REST_SERVICE_HOST, CODE_JAIL_REST_SERVICE_READ_TIMEOUT,
#   and CODE_JAIL_REST_SERVICE_CONNECT_TIMEOUT are configured.
# .. toggle_use_cases: temporary
# .. toggle_creation_date: 2025-04-03
# .. toggle_target_removal_date: 2025-05-01
ENABLE_CODEJAIL_DARKLAUNCH = SettingToggle("ENABLE_CODEJAIL_DARKLAUNCH", default=False, module_name=__name__)


def is_codejail_rest_service_enabled():
    """Return whether the codejail REST service is enabled."""
    return ENABLE_CODEJAIL_REST_SERVICE.is_enabled()


def is_codejail_in_darklaunch():
    """
    Returns whether codejail dark launch is enabled.

    Codejail dark launch can only be enabled if ENABLE_CODEJAIL_REST_SERVICE is not enabled.
    """
    return not is_codejail_rest_service_enabled() and ENABLE_CODEJAIL_DARKLAUNCH.is_enabled()


def get_remote_exec(*args, **kwargs):
    """Get remote exec function based on setting and executes it."""
    remote_exec_function_name = settings.CODE_JAIL_REST_SERVICE_REMOTE_EXEC
    try:
        mod_name, func_name = remote_exec_function_name.rsplit(".", 1)
        remote_exec_module = import_module(mod_name)
        remote_exec_function = getattr(remote_exec_module, func_name)
        if not remote_exec_function:
            remote_exec_function = send_safe_exec_request_v0
    except ModuleNotFoundError:
        return send_safe_exec_request_v0(*args, **kwargs)
    return remote_exec_function(*args, **kwargs)


def get_codejail_rest_service_endpoint():
    """Return the endpoint URL for the codejail REST service."""
    return f"{settings.CODE_JAIL_REST_SERVICE_HOST}/api/v0/code-exec"


def send_safe_exec_request_v0(data):
    """
    Sends a request to a codejail api service forwarding required code and files.
    Arguments:
        data: Dict containing code and other parameters required for jailed code execution.
        It also includes extra_files (python_lib.zip) required by the codejail execution.
    Returns:
        Response received from codejail api service
    """
    globals_dict = data["globals_dict"]
    extra_files = data.pop("extra_files")

    codejail_service_endpoint = get_codejail_rest_service_endpoint()

    # In rare cases an XBlock might introduce `bytes` objects (or other
    # non-JSON-serializable objects) into the globals dict. The codejail service
    # (via the codejail library) will call `json_safe` on the globals before
    # JSON-encoding for the sandbox input, but here we need to call it earlier
    # in the process so we can even transport the globals *to* the codejail
    # service. Otherwise, we may get a TypeError when constructing the payload.
    #
    # This is a lossy operation (non-serializable objects will be dropped, and
    # bytes converted to strings) but it is the same lossy operation that
    # codejail will perform anyhow -- and it should be idempotent.
    data_send = {**data}
    data_send["globals_dict"] = json_safe(data_send["globals_dict"])

    payload = json.dumps(data_send)

    try:
        response = requests.post(
            codejail_service_endpoint,
            files=extra_files,
            data={"payload": payload},
            timeout=(settings.CODE_JAIL_REST_SERVICE_CONNECT_TIMEOUT, settings.CODE_JAIL_REST_SERVICE_READ_TIMEOUT),
        )

    except RequestException as err:
        log.error(
            "Failed to connect to codejail api service: url=%s, params=%s", codejail_service_endpoint, str(payload)
        )
        raise CodejailServiceUnavailable(
            _("Codejail API Service is unavailable. Please try again in a few minutes.")
        ) from err

    try:
        response.raise_for_status()
    except HTTPError as err:
        raise CodejailServiceStatusError(_("Codejail API Service invalid response.")) from err

    try:
        response_json = response.json()
    except JSONDecodeError as err:
        log.error("Invalid JSON response received from codejail api service: Response_Content=%s", response.content)
        raise CodejailServiceParseError(_("Invalid JSON response received from codejail api service.")) from err

    emsg = response_json.get("emsg")
    exception = None

    if emsg:
        exception_msg = f"{emsg}. For more information check Codejail Service logs."
        exception = SafeExecException(exception_msg)

    globals_dict.update(response_json.get("globals_dict"))

    return emsg, exception



""" Mixin that provides AJAX handling for Video XBlock """
from webob import Response
from webob.multidict import MultiDict
from xblock.core import XBlock


class AjaxHandlerMixin:
    """
    Mixin that provides AJAX handling for Video XBlock
    """
    @property
    def ajax_url(self):
        """
        Returns the URL for the ajax handler.
        """
        return self.runtime.handler_url(self, 'ajax_handler', '', '').rstrip('/?')

    @XBlock.handler
    def ajax_handler(self, request, suffix=None):
        """
        XBlock handler that wraps `ajax_handler`
        """
        class FileObjForWebobFiles:
            """
            Turn Webob cgi.FieldStorage uploaded files into pure file objects.

            Webob represents uploaded files as cgi.FieldStorage objects, which
            have a .file attribute.  We wrap the FieldStorage object, delegating
            attribute access to the .file attribute.  But the files have no
            name, so we carry the FieldStorage .filename attribute as the .name.

            """
            def __init__(self, webob_file):
                self.file = webob_file.file
                self.name = webob_file.filename

            def __getattr__(self, name):
                return getattr(self.file, name)

        # WebOb requests have multiple entries for uploaded files.  handle_ajax
        # expects a single entry as a list.
        request_post = MultiDict(request.POST)
        for key in set(request.POST.keys()):
            if hasattr(request.POST[key], "file"):
                request_post[key] = list(map(FileObjForWebobFiles, request.POST.getall(key)))

        response_data = self.handle_ajax(suffix, request_post)
        return Response(response_data, content_type='application/json', charset='UTF-8')



"""pdfXBlock main Python class."""
import json

from django.utils.translation import gettext_noop as _
from web_fragments.fragment import Fragment
from webob import Response
from xblock.core import XBlock
from xblock.fields import Boolean, Scope, String
from xblock.utils.resources import ResourceLoader

from .utils import bool_from_str, is_all_download_disabled

resource_loader = ResourceLoader(__name__)


@XBlock.needs('i18n')
class PDFBlock(XBlock):
    """PDF XBlock. Allows authors to embed PDFs in their courses."""

    icon_class = "other"

    display_name = String(
        display_name=_("Display Name"),
        default=_("PDF"),
        scope=Scope.settings,
        help=_("This name appears in the horizontal navigation at the top of the page.")
    )

    url = String(
        display_name=_("PDF URL"),
        default=_("https://tutorial.math.lamar.edu/pdf/Trig_Cheat_Sheet.pdf"),
        scope=Scope.content,
        help=_("The URL for your PDF.")
    )

    allow_download = Boolean(
        display_name=_("PDF Download Allowed"),
        default=True,
        scope=Scope.content,
        help=_("Display a download button for this PDF.")
    )

    source_text = String(
        display_name=_("Source document button text"),
        default="",
        scope=Scope.content,
        help=_(
            "Add a download link for the source file of your PDF. "
            "Use it for example to provide the PowerPoint file used to create this PDF."
        )
    )

    source_url = String(
        display_name=_("Source document URL"),
        default="",
        scope=Scope.content,
        help=_(
            "Add a download link for the source file of your PDF. "
            "Use it for example to provide the PowerPoint file used to create this PDF."
        )
    )

    @property
    def raw_settings(self):
        """Get the raw settings of the XBlock as a dictionary."""
        return {
            'display_name': self.display_name,
            'url': self.url,
            'allow_download': self.allow_download,
            'disable_all_download': is_all_download_disabled(),
            'source_text': self.source_text,
            'source_url': self.source_url,
        }

    def student_view(self, context=None):  # pylint: disable=unused-argument
        """Primary view of the XBlock, shown to students when viewing courses."""
        html = resource_loader.render_django_template(
            'templates/html/pdf_view.html',
            context=self.raw_settings,
            i18n_service=self.runtime.service(self, "i18n"),
        )

        event_type = 'edx.pdf.loaded'
        event_data = {
            'url': self.url,
            'source_url': self.source_url,
        }
        self.runtime.publish(self, event_type, event_data)
        frag = Fragment(html)
        frag.add_javascript(resource_loader.load_unicode("static/js/pdf_view.js"))
        frag.initialize_js('pdfXBlockInitView')
        return frag

    def studio_view(self, context=None):
        """
        Secondary view of the XBlock.

        Shown to teachers when editing the XBlock.
        """
        context = {
            'display_name': self.display_name,
            'url': self.url,
            'allow_download': self.allow_download,
            'disable_all_download': is_all_download_disabled(),
            'source_text': self.source_text,
            'source_url': self.source_url
        }
        html = resource_loader.render_django_template(
            'templates/html/pdf_edit.html',
            context=context,
            i18n_service=self.runtime.service(self, "i18n"),
        )
        frag = Fragment(html)
        frag.add_javascript(resource_loader.load_unicode("static/js/pdf_edit.js"))
        frag.initialize_js('pdfXBlockInitEdit')
        return frag

    @XBlock.json_handler
    def on_download(self, data, suffix=''):  # pylint: disable=unused-argument
        """Download file event handler."""
        event_type = 'edx.pdf.downloaded'
        event_data = {
            'url': self.url,
            'source_url': self.source_url,
        }
        self.runtime.publish(self, event_type, event_data)

    @XBlock.handler
    def load_pdf(self, *_args, **_kwargs):
        """Get the PDF block's settings in JSON format."""
        return Response(json.dumps(self.raw_settings), content_type='application/json', charset='utf8')

    @XBlock.json_handler
    def save_pdf(self, data, suffix=''):  # pylint: disable=unused-argument
        """Save handler."""
        self.display_name = data['display_name']
        self.url = data['url']

        if not is_all_download_disabled():
            self.allow_download = bool_from_str(data['allow_download'])
            self.source_text = data['source_text']
            self.source_url = data['source_url']

        return {
            'result': 'success',
        }



"""
Constants used by DjangoXBlockUserService
"""

# This is the view that will be rendered to display the XBlock in the LMS.
# It will also be used to render the block in "preview" mode in Studio, unless
# the XBlock also implements author_view.
STUDENT_VIEW = 'student_view'

# This is the view that will be rendered to display the XBlock in the LMS for unenrolled learners.
# Implementations of this view should assume that a user and user data are not available.
PUBLIC_VIEW = 'public_view'

# The personally identifiable user ID.
ATTR_KEY_USER_ID = 'edx-platform.user_id'
# The country code determined from the user's request IP address.
ATTR_KEY_REQUEST_COUNTRY_CODE = 'edx-platform.request_country_code'



"""
Handlers for video block.

StudentViewHandlers are handlers for video block instance.
StudioViewHandlers are handlers for video descriptor instance.
"""


import json
import logging
import math

from django.utils.timezone import now
from opaque_keys.edx.locator import CourseLocator
from webob import Response
from xblock.core import XBlock
from xblock.exceptions import JsonHandlerError
from xblock.fields import RelativeTime

from xblocks_contrib.video.exceptions import TranscriptNotFoundError, TranscriptsGenerationException
from xblocks_contrib.video.video_transcripts_utils import TranscriptExtensions, clean_video_id, subs_filename
from xblocks_contrib.video.video_utils import load_metadata_from_youtube

log = logging.getLogger(__name__)


def get_transcript(
    video_block,
    lang: str | None = None,
    output_format: str = 'srt',
    youtube_id: str | None = None,
    is_bumper: bool = False,
) -> tuple[bytes, str, str]:
    """
    Retrieve a transcript using a video block's configuration service.

    Returns:
        tuple(bytes, str, str): transcript content, filename, and mimetype.

    Raises:
        Exception: If the video config service is not available or the transcript cannot be retrieved.
    """
    video_config_service = video_block.runtime.service(video_block, 'video_config')
    if not video_config_service:
        raise Exception("Video config service not found")
    return video_config_service.get_transcript(video_block, lang, output_format, youtube_id, is_bumper)


# Disable no-member warning:
# pylint: disable=no-member

def to_boolean(value):
    """
    Convert a value from a GET or POST request parameter to a bool
    """
    if isinstance(value, bytes):
        value = value.decode('ascii', errors='replace')
    if isinstance(value, str):
        return value.lower() == 'true'
    else:
        return bool(value)


class VideoStudentViewHandlers:
    """
    Handlers for video block instance.
    """
    global_speed = None
    transcript_language = None

    def handle_ajax(self, dispatch, data):
        """
        Update values of xfields, that were changed by student.
        """
        accepted_keys = [
            'speed', 'auto_advance', 'saved_video_position', 'transcript_language',
            'transcript_download_format', 'youtube_is_available',
            'bumper_last_view_date', 'bumper_do_not_show_again'
        ]

        conversions = {
            'speed': json.loads,
            'auto_advance': json.loads,
            'saved_video_position': RelativeTime.isotime_to_timedelta,
            'youtube_is_available': json.loads,
            'bumper_last_view_date': to_boolean,
            'bumper_do_not_show_again': to_boolean,
        }

        if dispatch == 'save_user_state':
            for key in data:
                if key in accepted_keys:
                    if key in conversions:
                        value = conversions[key](data[key])
                    else:
                        value = data[key]

                    if key == 'bumper_last_view_date':
                        value = now()

                    if key == 'speed' and math.isnan(value):
                        message = f"Invalid speed value {value}, must be a float."
                        log.warning(message)
                        return json.dumps({'success': False, 'error': message})

                    setattr(self, key, value)

                    if key == 'speed':
                        self.global_speed = self.speed

            return json.dumps({'success': True})

        log.debug(f"GET {data}")
        log.debug(f"DISPATCH {dispatch}")

        raise TranscriptNotFoundError('Unexpected dispatch type')

    def get_static_transcript(self, request, transcripts):
        """
        Courses that are imported with the --nostatic flag do not show
        transcripts/captions properly even if those captions are stored inside
        their static folder. This adds a last resort method of redirecting to
        the static asset path of the course if the transcript can't be found
        inside the contentstore and the course has the static_asset_path field
        set.

        transcripts (dict): A dict with all transcripts and a sub.
        """
        response = Response(status=404)
        # Only do redirect for English
        if not self.transcript_language == 'en':
            return response

        # If this video lives in library, the code below is not relevant and will error.
        if not isinstance(self.context_key, CourseLocator):
            return response

        video_id = request.GET.get('videoId', None)
        if video_id:
            transcript_name = video_id
        else:
            transcript_name = transcripts["sub"]

        if transcript_name:
            # Get the asset path for course
            asset_path = None
            course = self.runtime.modulestore.get_course(self.context_key)
            if course.static_asset_path:
                asset_path = course.static_asset_path
            else:
                # It seems static_asset_path is not set in any XMLModuleStore courses.
                asset_path = getattr(course, 'data_dir', '')

            if asset_path:
                response = Response(
                    status=307,
                    location='/static/{}/{}'.format(
                        asset_path,
                        subs_filename(transcript_name, self.transcript_language)
                    )
                )
        return response

    @XBlock.json_handler
    def publish_completion(self, data, dispatch):  # pylint: disable=unused-argument
        """
        Entry point for completion for student_view.

        Parameters:
            data: JSON dict:
                key: "completion"
                value: float in range [0.0, 1.0]

            dispatch: Ignored.
        Return value: JSON response (200 on success, 400 for malformed data)
        """
        completion_service = self.runtime.service(self, 'completion')
        if completion_service is None:
            raise JsonHandlerError(500, "No completion service found")
        if not completion_service.completion_tracking_enabled():
            raise JsonHandlerError(404, "Completion tracking is not enabled and API calls are unexpected")
        if not isinstance(data['completion'], (int, float)):
            message = "Invalid completion value {}. Must be a float in range [0.0, 1.0]"
            raise JsonHandlerError(400, message.format(data['completion']))
        if not 0.0 <= data['completion'] <= 1.0:
            message = "Invalid completion value {}. Must be in range [0.0, 1.0]"
            raise JsonHandlerError(400, message.format(data['completion']))
        self.runtime.publish(self, "completion", data)
        return {"result": "ok"}

    @staticmethod
    def make_transcript_http_response(content, filename, language, content_type, add_attachment_header=True):
        """
        Construct `Response` object.

        Arguments:
            content (unicode): transcript content
            filename (unicode): transcript filename
            language (unicode): transcript language
            mimetype (unicode): transcript content type
            add_attachment_header (bool): whether to add attachment header or not
        """
        headerlist = [
            ('Content-Language', language),
        ]

        if add_attachment_header:
            headerlist.append(
                (
                    'Content-Disposition',
                    f'attachment; filename="{filename}"'
                )
            )

        response = Response(
            content,
            headerlist=headerlist,
            charset='utf8'
        )
        response.content_type = content_type

        return response

    @XBlock.handler
    def transcript(self, request, dispatch):
        """
        Entry point for transcript handlers for student_view.

        Request GET contains::

            (optional) `videoId` for `translation` dispatch.
            `is_bumper=1` flag for bumper case.

        Dispatches, (HTTP GET)::

            /translation/[language_id]
            /download
            /available_translations/

        Explanations::

            `download`: returns SRT or TXT file.
            `translation`: depends on HTTP methods:
                    Provide translation for requested language, SJSON format is sent back on success,
                    Proper language_id should be in url.
            `available_translations`:
                    Returns list of languages, for which transcript files exist.
                    For 'en' check if SJSON exists. For non-'en' check if SRT file exists.
        """
        is_bumper = request.GET.get('is_bumper', False)
        transcripts = self.get_transcripts_info(is_bumper)

        if dispatch.startswith('translation'):
            language = dispatch.replace('translation', '').strip('/')

            # Because scrapers hit video blocks, verify that a user exists.
            # use the _request attr to get the django request object.
            if not request._request.user:  # pylint: disable=protected-access
                log.info("Transcript: user must be logged or public view enabled to get transcript")
                return Response(status=403)

            if not language:
                log.info("Invalid /translation request: no language.")
                return Response(status=400)

            if language not in ['en'] + list(transcripts["transcripts"].keys()):
                log.info("Video: transcript facilities are not available for given language.")
                return Response(status=404)

            if language != self.transcript_language:
                self.transcript_language = language

            try:
                youtube_id = None if is_bumper else request.GET.get('videoId')
                content, filename, mimetype = get_transcript(
                    self,
                    lang=self.transcript_language,
                    output_format=TranscriptExtensions.SJSON,
                    youtube_id=youtube_id,
                    is_bumper=is_bumper
                )
                response = self.make_transcript_http_response(
                    content,
                    filename,
                    self.transcript_language,
                    mimetype,
                    add_attachment_header=False
                )
            except TranscriptNotFoundError as exc:
                edx_video_id = clean_video_id(self.edx_video_id)
                log.warning(
                    '[Translation Dispatch] %s: %s',
                    self.usage_key,
                    exc if is_bumper else f'Transcript not found for {edx_video_id}, lang: {self.transcript_language}',
                )
                response = self.get_static_transcript(request, transcripts)

        elif dispatch == 'download':
            lang = request.GET.get('lang', None)

            try:
                content, filename, mimetype = get_transcript(self, lang, output_format=self.transcript_download_format)
            except TranscriptNotFoundError:
                return Response(status=404)

            response = self.make_transcript_http_response(
                content,
                filename,
                self.transcript_language,
                mimetype
            )
        elif dispatch.startswith('available_translations'):
            video_config_service = self.runtime.service(self, 'video_config')
            if not video_config_service:
                return Response(status=404)
            available_translations = video_config_service.available_translations(
                self,
                transcripts,
                verify_assets=True,
                is_bumper=is_bumper
            )
            if available_translations:
                response = Response(json.dumps(available_translations))
                response.content_type = 'application/json'
            else:
                response = Response(status=404)
        else:  # unknown dispatch
            log.debug("Dispatch is not allowed")
            response = Response(status=404)

        return response

    @XBlock.handler
    def student_view_user_state(self, request, suffix=''):  # lint-amnesty, pylint: disable=unused-argument
        """
        Endpoint to get user-specific state, like current position and playback speed,
        without rendering the full student_view HTML. This is similar to student_view_state,
        but that one cannot contain user-specific info.
        """
        view_state = self.student_view_data()
        view_state.update({
            "saved_video_position": self.saved_video_position.total_seconds(),
            "speed": self.speed,
        })
        return Response(
            json.dumps(view_state),
            content_type='application/json',
            charset='UTF-8'
        )

    @XBlock.handler
    def yt_video_metadata(self, request, suffix=''):  # lint-amnesty, pylint: disable=unused-argument
        """
        Endpoint to get YouTube metadata.
        This handler is only used in the openedx_content-based runtime. The old
        runtime uses a similar REST API that's not an XBlock handler.
        """
        if not self.youtube_id_1_0:
            # TODO: more informational response to explain that yt_video_metadata not supported for non-youtube videos.
            return Response('{}', status=400)

        metadata, status_code = load_metadata_from_youtube(video_id=self.youtube_id_1_0, request=request)
        response = Response(json.dumps(metadata), status=status_code)
        response.content_type = 'application/json'
        return response


class VideoStudioViewHandlers:
    """
    Handlers for Studio view.
    """
    def validate_transcript_upload_data(self, data):
        """
        Validates video transcript file.
        Arguments:
            data: Transcript data to be validated.
        Returns:
            None or String
            If there is error returns error message otherwise None.
        """
        error = None
        _ = self.runtime.service(self, "i18n").ugettext
        # Validate the must have attributes - this error is unlikely to be faced by common users.
        must_have_attrs = ['edx_video_id', 'language_code', 'new_language_code']
        missing = [attr for attr in must_have_attrs if attr not in data]

        # Get available transcript languages.
        transcripts = self.get_transcripts_info()
        video_config_service = self.runtime.service(self, 'video_config')
        if not video_config_service:
            return error
        available_translations = video_config_service.available_translations(
            self,
            transcripts,
            verify_assets=True
        )

        if missing:
            error = _('The following parameters are required: {missing}.').format(missing=', '.join(missing))
        elif (
            data['language_code'] != data['new_language_code'] and data['new_language_code'] in available_translations
        ):
            error = _('A transcript with the "{language_code}" language code already exists.').format(
                language_code=data['new_language_code'],
            )
        elif 'file' not in data:
            error = _('A transcript file is required.')

        return error

    @XBlock.handler
    def studio_transcript(self, request, dispatch):
        """
        Entry point for Studio transcript handlers.

        Dispatches::

            /translation/[language_id] - language_id should be in url.

        ``translation`` dispatch support following HTTP methods::

            `POST`:
                Upload srt file. Check possibility of generation of proper sjson files.
                For now, it works only for self.transcripts, not for `en`.
            `GET`:
                Return filename from storage. SRT format is sent back on success. Filename should be in GET dict.

        We raise all exceptions right in Studio::

            TranscriptNotFoundError:
                Video or asset was deleted from module/contentstore, but request came later.
                Seems impossible to be raised. block_render.py catches NotFoundErrors from here.

            /translation POST:
                TypeError:
                    Unjsonable filename or content.
                TranscriptsGenerationException, TranscriptException:
                    no SRT extension or not parse-able by PySRT
                UnicodeDecodeError: non-UTF8 uploaded file content encoding.
        """
        if dispatch.startswith('translation'):

            if request.method == 'POST':
                response = self._studio_transcript_upload(request)
            elif request.method == 'DELETE':
                response = self._studio_transcript_delete(request)
            elif request.method == 'GET':
                response = self._studio_transcript_get(request)
            else:
                # Any other HTTP method is not allowed.
                response = Response(status=404)

        else:  # unknown dispatch
            log.debug("Dispatch is not allowed")
            response = Response(status=404)

        return response

    def _studio_transcript_upload(self, request):
        """
        Upload transcript. Used in "POST" method in `studio_transcript`
        """
        _ = self.runtime.service(self, "i18n").ugettext
        video_config_service = self.runtime.service(self, 'video_config')
        if not video_config_service:
            return Response(json={'error': _('Runtime does not support transcripts.')}, status=400)
        error = self.validate_transcript_upload_data(data=request.POST)
        if error:
            return Response(json={'error': error}, status=400)
        edx_video_id = (request.POST['edx_video_id'] or "").strip()
        language_code = request.POST['language_code']
        new_language_code = request.POST['new_language_code']
        try:
            video_config_service.upload_transcript(
                video_block=self,  # NOTE: .edx_video_id and .transcripts may get mutated
                edx_video_id=edx_video_id,
                language_code=language_code,
                new_language_code=new_language_code,
                transcript_file=request.POST['file'].file,
            )
            return Response(
                json.dumps(
                    {
                        "edx_video_id": edx_video_id or self.edx_video_id,
                        "language_code": new_language_code,
                    }
                ),
                status=201,
            )
        except (TranscriptsGenerationException, UnicodeDecodeError):
            return Response(
                json={
                    'error': _(
                        'There is a problem with this transcript file. Try to upload a different file.'
                    )
                },
                status=400
            )

    def _studio_transcript_delete(self, request):
        """
        Delete transcript. Used in "DELETE" method in `studio_transcript`
        """
        _ = self.runtime.service(self, "i18n").ugettext
        video_config_service = self.runtime.service(self, 'video_config')
        if not video_config_service:
            return Response(json={'error': _('Runtime does not support transcripts.')}, status=400)
        request_data = request.json
        if 'lang' not in request_data or 'edx_video_id' not in request_data:
            return Response(status=400)
        video_config_service.delete_transcript(
            video_block=self,
            edx_video_id=request_data['edx_video_id'],
            language_code=request_data['lang'],
        )
        return Response(status=200)

    def _studio_transcript_get(self, request):
        """
        Get transcript. Used in "GET" method in `studio_transcript`
        """
        _ = self.runtime.service(self, "i18n").ugettext
        language = request.GET.get('language_code')
        if not language:
            return Response(json={'error': _('Language is required.')}, status=400)

        try:
            video_config_service = self.runtime.service(self, 'video_config')
            if not video_config_service:
                return Response(status=404)
            transcript_content, transcript_name, mime_type = video_config_service.get_transcript(
                self, lang=language, output_format=TranscriptExtensions.SRT
            )
            response = Response(transcript_content, headerlist=[
                (
                    'Content-Disposition',
                    f'attachment; filename="{transcript_name}"'
                ),
                ('Content-Language', language),
                ('Content-Type', mime_type)
            ])
        except (
            UnicodeDecodeError,
            TranscriptsGenerationException,
            TranscriptNotFoundError
        ):
            response = Response(status=404)
        return response



"""
Utils for video bumper
"""


import copy
import json
import logging
from collections import OrderedDict
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

from django.conf import settings

from .video_utils import get_edxval_api, set_query_parameter

log = logging.getLogger(__name__)


def get_bumper_settings(video):
    """
    Get bumper settings from video instance.
    """
    bumper_settings = copy.deepcopy(getattr(video, 'video_bumper', {}))

    # clean up /static/ prefix from bumper transcripts
    for lang, transcript_url in bumper_settings.get('transcripts', {}).items():
        bumper_settings['transcripts'][lang] = transcript_url.replace("/static/", "")

    return bumper_settings


def is_bumper_enabled(video):
    """
    Check if bumper enabled.

    - Feature flag ENABLE_VIDEO_BUMPER should be set to True
    - Do not show again button should not be clicked by user.
    - Current time minus periodicity must be greater that last time viewed
    - edxval_api should be presented

    Returns:
         bool.
    """
    bumper_last_view_date = getattr(video, 'bumper_last_view_date', None)
    utc_now = datetime.now(ZoneInfo("UTC"))
    periodicity = settings.FEATURES.get('SHOW_BUMPER_PERIODICITY', 0)
    has_viewed = any([
        video.bumper_do_not_show_again,
        (bumper_last_view_date and bumper_last_view_date + timedelta(seconds=periodicity) > utc_now)
    ])
    is_studio = getattr(video.runtime, "is_author_mode", False)
    edxval_api = get_edxval_api()
    return bool(
        not is_studio and
        settings.FEATURES.get('ENABLE_VIDEO_BUMPER') and
        get_bumper_settings(video) and
        edxval_api and
        not has_viewed
    )


def bumperize(video):
    """
    Populate video with bumper settings, if they are presented.
    """
    video.bumper = {
        'enabled': False,
        'edx_video_id': "",
        'transcripts': {},
        'metadata': None,
    }

    if not is_bumper_enabled(video):
        return

    bumper_settings = get_bumper_settings(video)

    try:
        video.bumper['edx_video_id'] = bumper_settings['video_id']
        video.bumper['transcripts'] = bumper_settings['transcripts']
    except (TypeError, KeyError):
        log.warning(
            "Could not retrieve video bumper information from course settings"
        )
        return

    sources = get_bumper_sources(video)
    if not sources:
        return

    video.bumper.update({
        'metadata': bumper_metadata(video, sources),
        'enabled': True,  # Video poster needs this.
    })


def get_bumper_sources(video):
    """
    Get bumper sources from edxval.

    Returns list of sources.
    """
    edxval_api = get_edxval_api()
    try:
        val_profiles = ["desktop_webm", "desktop_mp4"]
        val_video_urls = edxval_api.get_urls_for_profiles(video.bumper['edx_video_id'], val_profiles)
        bumper_sources = [url for url in [val_video_urls[p] for p in val_profiles] if url]
    except edxval_api.ValInternalError:
        # if no bumper sources, nothing will be showed
        log.warning(
            "Could not retrieve information from VAL for Bumper edx Video ID: %s.", video.bumper['edx_video_id']
        )
        return []

    return bumper_sources


def bumper_metadata(video, sources):
    """
    Generate bumper metadata.
    """
    transcripts = video.get_transcripts_info(is_bumper=True)
    unused_track_url, bumper_transcript_language, bumper_languages = video.get_transcripts_for_student(transcripts)

    metadata = OrderedDict({
        'saveStateUrl': video.ajax_url + '/save_user_state',
        'showCaptions': json.dumps(video.show_captions),
        'sources': sources,
        'streams': '',
        'transcriptLanguage': bumper_transcript_language,
        'transcriptLanguages': bumper_languages,
        'transcriptTranslationUrl': set_query_parameter(
            video.runtime.handler_url(video, 'transcript', 'translation/__lang__').rstrip('/?'), 'is_bumper', 1
        ),
        'transcriptAvailableTranslationsUrl': set_query_parameter(
            video.runtime.handler_url(video, 'transcript', 'available_translations').rstrip('/?'), 'is_bumper', 1
        ),
        'publishCompletionUrl': set_query_parameter(
            video.runtime.handler_url(video, 'publish_completion', '').rstrip('?'), 'is_bumper', 1
        ),
    })

    return metadata



"""Factories to build CAPA response XML."""

from abc import ABCMeta, abstractmethod

import six
from lxml import etree
from six.moves import range, zip


class ResponseXMLFactory(six.with_metaclass(ABCMeta, object)):
    """Abstract base class for capa response XML factories.
    Subclasses override create_response_element and
    create_input_element to produce XML of particular response types"""

    @abstractmethod
    def create_response_element(self, **kwargs):
        """Subclasses override to return an etree element
        representing the capa response XML
        (e.g. <numericalresponse>).

        The tree should NOT contain any input elements
        (such as <textline />) as these will be added later."""
        return None

    @abstractmethod
    def create_input_element(self, **kwargs):
        """Subclasses override this to return an etree element
        representing the capa input XML (such as <textline />)"""
        return None

    def build_xml(self, **kwargs):  # pylint: disable=too-many-locals
        """Construct an XML string for a capa response
        based on ``**kwargs``.

        ``**kwargs`` is a dictionary that will be passed
        to create_response_element() and create_input_element().
        See the subclasses below for other keyword arguments
        you can specify.

        For all response types, ``**kwargs`` can contain:

        *question_text*: The text of the question to display,
            wrapped in <label> tags.

        *explanation_text*: The detailed explanation that will
            be shown if the user answers incorrectly.

        *script*: The embedded Python script (a string)

        *num_responses*: The number of responses to create [DEFAULT: 1]

        *num_inputs*: The number of input elements
            to create [DEFAULT: 1]

        *credit_type*: String of comma-separated words specifying the
            partial credit grading scheme.

        Returns a string representation of the XML tree.
        """

        # Retrieve keyward arguments
        question_text = kwargs.get("question_text", "")
        explanation_text = kwargs.get("explanation_text", "")
        script = kwargs.get("script", None)
        num_responses = kwargs.get("num_responses", 1)
        num_inputs = kwargs.get("num_inputs", 1)
        credit_type = kwargs.get("credit_type", None)

        # The root is <problem>
        root = etree.Element("problem")

        # Add a script if there is one
        if script:
            script_element = etree.SubElement(root, "script")
            script_element.set("type", "loncapa/python")
            script_element.text = str(script)

        # Add the response(s)
        for __ in range(int(num_responses)):
            response_element = self.create_response_element(**kwargs)

            # Set partial credit
            if credit_type is not None:
                response_element.set("partial_credit", str(credit_type))

            root.append(response_element)

            # Add the question label
            question = etree.SubElement(response_element, "label")
            question.text = question_text

            # Add input elements
            for __ in range(int(num_inputs)):
                input_element = self.create_input_element(**kwargs)
                if input_element is not None:
                    response_element.append(input_element)

            # The problem has an explanation of the solution
            if explanation_text:
                explanation = etree.SubElement(root, "solution")
                explanation_div = etree.SubElement(explanation, "div")
                explanation_div.set("class", "detailed-solution")
                explanation_div.text = explanation_text

        return etree.tostring(root).decode("utf-8")

    @staticmethod
    def textline_input_xml(**kwargs):
        """Create a <textline/> XML element

        Uses ``**kwargs``:

        *math_display*: If True, then includes a MathJax display of user input

        *size*: An integer representing the width of the text line
        """
        math_display = kwargs.get("math_display", False)
        size = kwargs.get("size", None)
        input_element_label = kwargs.get("input_element_label", "")

        input_element = etree.Element("textline")

        if input_element_label:
            input_element.set("label", input_element_label)

        if math_display:
            input_element.set("math", "1")

        if size:
            input_element.set("size", str(size))

        return input_element

    @staticmethod
    def choicegroup_input_xml(**kwargs):
        """Create a <choicegroup> XML element

        Uses ``**kwargs``:

        *choice_type*: Can be "checkbox", "radio", or "multiple"

        *choices*: List of True/False values indicating whether
                            a particular choice is correct or not.
                            Users must choose *all* correct options in order
                            to be marked correct.
                            DEFAULT: [True]

        *choice_names*: List of strings identifying the choices.
                        If specified, you must ensure that
                        len(choice_names) == len(choices)

        *points*: List of strings giving partial credit values (0-1)
                  for each choice. Interpreted as floats in problem.
                  If specified, ensure len(points) == len(choices)
        """
        # Names of group elements
        group_element_names = {"checkbox": "checkboxgroup", "radio": "radiogroup", "multiple": "choicegroup"}

        # Retrieve **kwargs
        choices = kwargs.get("choices", [True])
        choice_type = kwargs.get("choice_type", "multiple")
        choice_names = kwargs.get("choice_names", [None] * len(choices))
        points = kwargs.get("points", [None] * len(choices))

        # Create the <choicegroup>, <checkboxgroup>, or <radiogroup> element
        assert choice_type in group_element_names
        group_element = etree.Element(group_element_names[choice_type])

        # Create the <choice> elements
        for correct_val, name, pointval in zip(choices, choice_names, points):
            choice_element = etree.SubElement(group_element, "choice")
            if correct_val is True:
                correctness = "true"
            elif correct_val is False:
                correctness = "false"
            elif "partial" in correct_val:
                correctness = "partial"
            else:
                correctness = correct_val

            choice_element.set("correct", correctness)

            # Add a name identifying the choice, if one exists
            # For simplicity, we use the same string as both the
            # name attribute and the text of the element
            if name:
                choice_element.text = str(name)
                choice_element.set("name", str(name))

            # Add point values for partially-correct choices.
            if pointval:
                choice_element.set("point_value", str(pointval))

        return group_element


class NumericalResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <numericalresponse> XML trees"""

    def create_response_element(self, **kwargs):
        """Create a <numericalresponse> XML element.
        Uses ``**kwarg`` keys:

        *answer*: The correct answer (e.g. "5")

        *correcthint*: The feedback describing correct answer.

        *additional_answers*: A dict of additional answers along with their correcthint.

        *tolerance*: The tolerance within which a response
        is considered correct.  Can be a decimal (e.g. "0.01")
        or percentage (e.g. "2%")

        *credit_type*: String of comma-separated words specifying the
        partial credit grading scheme.

        *partial_range*: The multiplier for the tolerance that will
        still provide partial credit in the "close" grading style

        *partial_answers*: A string of comma-separated alternate
        answers that will receive partial credit in the "list" style
        """

        answer = kwargs.get("answer", None)
        correcthint = kwargs.get("correcthint", "")
        additional_answers = kwargs.get("additional_answers", {})
        tolerance = kwargs.get("tolerance", None)
        credit_type = kwargs.get("credit_type", None)
        partial_range = kwargs.get("partial_range", None)
        partial_answers = kwargs.get("partial_answers", None)

        response_element = etree.Element("numericalresponse")

        if answer:
            if isinstance(answer, float):
                response_element.set("answer", repr(answer))
            else:
                response_element.set("answer", str(answer))

        for additional_answer, additional_correcthint in additional_answers.items():
            additional_element = etree.SubElement(response_element, "additional_answer")
            additional_element.set("answer", str(additional_answer))
            if additional_correcthint:
                correcthint_element = etree.SubElement(additional_element, "correcthint")
                correcthint_element.text = str(additional_correcthint)

        if tolerance:
            responseparam_element = etree.SubElement(response_element, "responseparam")
            responseparam_element.set("type", "tolerance")
            responseparam_element.set("default", str(tolerance))
            if partial_range is not None and "close" in credit_type:
                responseparam_element.set("partial_range", str(partial_range))

        if partial_answers is not None and "list" in credit_type:
            # The line below throws a false positive pylint violation, so it's excepted.
            responseparam_element = etree.SubElement(response_element, "responseparam")
            responseparam_element.set("partial_answers", partial_answers)

        if correcthint:
            correcthint_element = etree.SubElement(response_element, "correcthint")
            correcthint_element.text = str(correcthint)

        return response_element

    def create_input_element(self, **kwargs):
        return ResponseXMLFactory.textline_input_xml(**kwargs)


class CustomResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <customresponse> XML trees"""

    def create_response_element(self, **kwargs):
        """Create a <customresponse> XML element.
        Uses ``**kwargs``:

        *cfn*: the Python code to run.  Can be inline code, or the name of a function defined in earlier <script> tags.

        Should have the form: cfn(expect, answer_given, student_answers)
        where expect is a value (see below),
        answer_given is a single value (for 1 input)
        or a list of values (for multiple inputs),
        and student_answers is a dict of answers by input ID.

        *expect*: The value passed to the function cfn

        *answer*: Inline script that calculates the answer

        *answer_attr*: The "answer" attribute on the tag itself (treated as an
        alias to "expect", though "expect" takes priority if both are given)
        """

        # Retrieve **kwargs
        cfn = kwargs.get("cfn", None)
        expect = kwargs.get("expect", None)
        answer_attr = kwargs.get("answer_attr", None)
        answer = kwargs.get("answer", None)
        options = kwargs.get("options", None)
        cfn_extra_args = kwargs.get("cfn_extra_args", None)

        # Create the response element
        response_element = etree.Element("customresponse")

        if cfn:
            response_element.set("cfn", str(cfn))

        if expect:
            response_element.set("expect", str(expect))

        if answer_attr:
            response_element.set("answer", str(answer_attr))

        if answer:
            answer_element = etree.SubElement(response_element, "answer")
            answer_element.text = str(answer)

        if options:
            response_element.set("options", str(options))

        if cfn_extra_args:
            response_element.set("cfn_extra_args", str(cfn_extra_args))

        return response_element

    def create_input_element(self, **kwargs):
        return ResponseXMLFactory.textline_input_xml(**kwargs)


class SchematicResponseXMLFactory(ResponseXMLFactory):
    """Factory for creating <schematicresponse> XML trees"""

    def create_response_element(self, **kwargs):
        """Create the <schematicresponse> XML element.

        Uses *kwargs*:

        *answer*: The Python script used to evaluate the answer.
        """
        answer_script = kwargs.get("answer", None)

        # Create the <schematicresponse> element
        response_element = etree.Element("schematicresponse")

        # Insert the <answer> script if one is provided
        if answer_script:
            answer_element = etree.SubElement(response_element, "answer")
            answer_element.set("type", "loncapa/python")
            answer_element.text = str(answer_script)

        return response_element

    def create_input_element(self, **kwargs):
        """Create the <schematic> XML element.

        Although <schematic> can have several attributes,
        (*height*, *width*, *parts*, *analyses*, *submit_analysis*, and *initial_value*),
        none of them are used in the capa block.
        For testing, we create a bare-bones version of <schematic>."""
        return etree.Element("schematic")


class CodeResponseXMLFactory(ResponseXMLFactory):
    """Factory for creating <coderesponse> XML trees"""

    def build_xml(self, **kwargs):
        """Build a <coderesponse> XML tree."""
        # Since we are providing an <answer> tag,
        # we should override the default behavior
        # of including a <solution> tag as well
        kwargs["explanation_text"] = None
        return super().build_xml(**kwargs)

    def create_response_element(self, **kwargs):
        """
        Create a <coderesponse> XML element.

        Uses ``**kwargs``:

        *initial_display*: The code that initially appears in the textbox
                            [DEFAULT: "Enter code here"]
        *answer_display*: The answer to display to the student
                            [DEFAULT: "This is the correct answer!"]
        *grader_payload*: A JSON-encoded string sent to the grader
                            [DEFAULT: empty dict string]
        *allowed_files*: A space-separated string of file names.
                            [DEFAULT: None]
        *required_files*: A space-separated string of file names.
                            [DEFAULT: None]

        """
        # Get **kwargs
        initial_display = kwargs.get("initial_display", "Enter code here")
        answer_display = kwargs.get("answer_display", "This is the correct answer!")
        grader_payload = kwargs.get("grader_payload", "{}")
        allowed_files = kwargs.get("allowed_files", None)
        required_files = kwargs.get("required_files", None)

        # Create the <coderesponse> element
        response_element = etree.Element("coderesponse")

        # If files are involved, create the <filesubmission> element.
        has_files = allowed_files or required_files
        if has_files:
            filesubmission_element = etree.SubElement(response_element, "filesubmission")
            if allowed_files:
                filesubmission_element.set("allowed_files", allowed_files)
            if required_files:
                filesubmission_element.set("required_files", required_files)

        # Create the <codeparam> element.
        codeparam_element = etree.SubElement(response_element, "codeparam")

        # Set the initial display text
        initial_element = etree.SubElement(codeparam_element, "initial_display")
        initial_element.text = str(initial_display)

        # Set the answer display text
        answer_element = etree.SubElement(codeparam_element, "answer_display")
        answer_element.text = str(answer_display)

        # Set the grader payload string
        grader_element = etree.SubElement(codeparam_element, "grader_payload")
        grader_element.text = str(grader_payload)

        # Create the input within the response
        if not has_files:
            input_element = etree.SubElement(response_element, "textbox")
            input_element.set("mode", "python")

        return response_element

    def create_input_element(self, **kwargs):
        # Since we create this in create_response_element(),
        # return None here
        return None


class ChoiceResponseXMLFactory(ResponseXMLFactory):
    """Factory for creating <choiceresponse> XML trees"""

    def create_response_element(self, **kwargs):
        """Create a <choiceresponse> element"""
        return etree.Element("choiceresponse")

    def create_input_element(self, **kwargs):
        """Create a <checkboxgroup> element."""
        return ResponseXMLFactory.choicegroup_input_xml(**kwargs)


class FormulaResponseXMLFactory(ResponseXMLFactory):
    """Factory for creating <formularesponse> XML trees"""

    def create_response_element(self, **kwargs):  # pylint: disable=too-many-locals
        """Create a <formularesponse> element.

        *sample_dict*: A dictionary of the form:
                        { VARIABLE_NAME: (MIN, MAX), ....}

                        This specifies the range within which
                        to numerically sample each variable to check
                        student answers.
                        [REQUIRED]

        *num_samples*: The number of times to sample the student's answer
                        to numerically compare it to the correct answer.

        *tolerance*: The tolerance within which answers will be accepted
                        [DEFAULT: 0.01]

        *answer*: The answer to the problem.  Can be a formula string
                    or a Python variable defined in a script
                    (e.g. "$calculated_answer" for a Python variable
                    called calculated_answer)
                    [REQUIRED]

        *hints*: List of (hint_prompt, hint_name, hint_text) tuples
                Where *hint_prompt* is the formula for which we show the hint,
                *hint_name* is an internal identifier for the hint,
                and *hint_text* is the text we show for the hint.
        """
        # Retrieve kwargs
        sample_dict = kwargs.get("sample_dict", None)
        num_samples = kwargs.get("num_samples", None)
        tolerance = kwargs.get("tolerance", 0.01)
        answer = kwargs.get("answer", None)
        hint_list = kwargs.get("hints", None)

        assert answer
        assert sample_dict and num_samples

        # Create the <formularesponse> element
        response_element = etree.Element("formularesponse")

        # Set the sample information
        sample_str = self._sample_str(sample_dict, num_samples, tolerance)
        response_element.set("samples", sample_str)

        # Set the tolerance
        responseparam_element = etree.SubElement(response_element, "responseparam")
        responseparam_element.set("type", "tolerance")
        responseparam_element.set("default", str(tolerance))

        # Set the answer
        response_element.set("answer", str(answer))

        # Include hints, if specified
        if hint_list:
            hintgroup_element = etree.SubElement(response_element, "hintgroup")

            for hint_prompt, hint_name, hint_text in hint_list:

                # For each hint, create a <formulahint> element
                formulahint_element = etree.SubElement(hintgroup_element, "formulahint")

                # We could sample a different range, but for simplicity,
                # we use the same sample string for the hints
                # that we used previously.
                formulahint_element.set("samples", sample_str)

                formulahint_element.set("answer", str(hint_prompt))
                formulahint_element.set("name", str(hint_name))

                # For each hint, create a <hintpart> element
                # corresponding to the <formulahint>
                hintpart_element = etree.SubElement(hintgroup_element, "hintpart")
                hintpart_element.set("on", str(hint_name))
                text_element = etree.SubElement(hintpart_element, "text")
                text_element.text = str(hint_text)

        return response_element

    def create_input_element(self, **kwargs):
        return ResponseXMLFactory.textline_input_xml(**kwargs)

    def _sample_str(self, sample_dict, num_samples, tolerance):  # pylint: disable=unused-argument
        """Generate a sample string for Loncapa using variable ranges and repetition count."""

        # Loncapa uses a special format for sample strings:
        # "x,y,z@4,5,3:10,12,8#4" means plug in values for (x,y,z)
        # from within the box defined by points (4,5,3) and (10,12,8)
        # The "#4" means to repeat 4 times.
        low_range_vals = [str(f[0]) for f in sample_dict.values()]
        high_range_vals = [str(f[1]) for f in sample_dict.values()]
        sample_str = (
            ",".join(list(sample_dict.keys()))
            + "@"
            + ",".join(low_range_vals)
            + ":"
            + ",".join(high_range_vals)
            + "#"
            + str(num_samples)
        )
        return sample_str


class ImageResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <imageresponse> XML"""

    def create_response_element(self, **kwargs):
        """Create the <imageresponse> element."""
        return etree.Element("imageresponse")

    def create_input_element(self, **kwargs):
        """Create the <imageinput> element.

        Uses ``**kwargs``:

        *src*: URL for the image file [DEFAULT: "/static/image.jpg"]

        *width*: Width of the image [DEFAULT: 100]

        *height*: Height of the image [DEFAULT: 100]

        *rectangle*: String representing the rectangles the user should select.

                    Take the form "(x1,y1)-(x2,y2)", where the two (x,y)
                    tuples define the corners of the rectangle.

                    Can include multiple rectangles separated by a semicolon, e.g.
                    "(490,11)-(556,98);(242,202)-(296,276)"

        *regions*: String representing the regions a user can select

                    Take the form "[ [[x1,y1], [x2,y2], [x3,y3]], [[x1,y1], [x2,y2], [x3,y3]] ]"
                    (Defines two regions, each with 3 points)

        REQUIRED: Either *rectangle* or *region* (or both)
        """

        # Get the **kwargs
        src = kwargs.get("src", "/static/image.jpg")
        width = kwargs.get("width", 100)
        height = kwargs.get("height", 100)
        rectangle = kwargs.get("rectangle", None)
        regions = kwargs.get("regions", None)

        assert rectangle or regions

        # Create the <imageinput> element
        input_element = etree.Element("imageinput")
        input_element.set("src", str(src))
        input_element.set("width", str(width))
        input_element.set("height", str(height))

        if rectangle:
            input_element.set("rectangle", rectangle)

        if regions:
            input_element.set("regions", regions)

        return input_element


class JSInputXMLFactory(CustomResponseXMLFactory):
    """
    Factory for producing <jsinput> XML.
    Note that this factory currently does not create a functioning problem.
    It will only create an empty iframe.
    """

    def create_input_element(self, **kwargs):
        """Create the <jsinput> element"""
        return etree.Element("jsinput")


class MultipleChoiceResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <multiplechoiceresponse> XML"""

    def create_response_element(self, **kwargs):
        """Create the <multiplechoiceresponse> element"""
        return etree.Element("multiplechoiceresponse")

    def create_input_element(self, **kwargs):
        """Create the <choicegroup> element"""
        kwargs["choice_type"] = "multiple"
        return ResponseXMLFactory.choicegroup_input_xml(**kwargs)


class TrueFalseResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <truefalseresponse> XML"""

    def create_response_element(self, **kwargs):
        """Create the <truefalseresponse> element"""
        return etree.Element("truefalseresponse")

    def create_input_element(self, **kwargs):
        """Create the <choicegroup> element"""
        kwargs["choice_type"] = "multiple"
        return ResponseXMLFactory.choicegroup_input_xml(**kwargs)


class OptionResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <optionresponse> XML"""

    def create_response_element(self, **kwargs):
        """Create the <optionresponse> element"""
        return etree.Element("optionresponse")

    def create_input_element(self, **kwargs):
        """Create the <optioninput> element.

        Uses ``**kwargs``:

        *options*: a list of possible options the user can choose from [REQUIRED] You must specify at least 2 options.
        *correct_option*: the correct choice from the list of options [REQUIRED]
        """

        options_list = kwargs.get("options", None)
        correct_option = kwargs.get("correct_option", None)

        assert options_list and correct_option
        assert len(options_list) > 1
        assert correct_option in options_list

        # Create the <optioninput> element
        optioninput_element = etree.Element("optioninput")

        # Set the "options" attribute
        # Format: "('first', 'second', 'third')"
        options_attr_string = ",".join([f"'{o}'" for o in options_list])
        options_attr_string = f"({options_attr_string})"
        optioninput_element.set("options", options_attr_string)

        # Set the "correct" attribute
        optioninput_element.set("correct", str(correct_option))

        return optioninput_element


class StringResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <stringresponse> XML"""

    def create_response_element(self, **kwargs):  # pylint: disable=too-many-locals
        """Create a <stringresponse> XML element.

        Uses ``**kwargs``:

        *answer*: The correct answer (a string) [REQUIRED]

        *case_sensitive*: Whether the response is case-sensitive (True/False)
                        [DEFAULT: True]

        *hints*: List of (hint_prompt, hint_name, hint_text) tuples
            Where *hint_prompt* is the string for which we show the hint,
            *hint_name* is an internal identifier for the hint,
            and *hint_text* is the text we show for the hint.

        *hintfn*: The name of a function in the script to use for hints.

        *regexp*: Whether the response is regexp

        *additional_answers*: list of additional answers.

        *non_attribute_answers*: list of additional answers to be coded in the
            non-attribute format

        """
        # Retrieve the **kwargs
        answer = kwargs.get("answer", None)
        case_sensitive = kwargs.get("case_sensitive", None)
        hint_list = kwargs.get("hints", None)
        hint_fn = kwargs.get("hintfn", None)
        regexp = kwargs.get("regexp", None)
        additional_answers = kwargs.get("additional_answers", [])
        non_attribute_answers = kwargs.get("non_attribute_answers", [])
        assert answer

        # Create the <stringresponse> element
        response_element = etree.Element("stringresponse")

        # Set the answer attribute
        response_element.set("answer", str(answer))

        # Set the case sensitivity and regexp:
        type_value = ""
        if case_sensitive is not None:
            type_value += "cs" if case_sensitive else "ci"
        type_value += " regexp" if regexp else ""
        if type_value:
            response_element.set("type", type_value.strip())

        # Add the hints if specified
        if hint_list or hint_fn:
            hintgroup_element = etree.SubElement(response_element, "hintgroup")
            if hint_list:
                assert not hint_fn
                for hint_prompt, hint_name, hint_text in hint_list:
                    stringhint_element = etree.SubElement(hintgroup_element, "stringhint")
                    stringhint_element.set("answer", str(hint_prompt))
                    stringhint_element.set("name", str(hint_name))

                    hintpart_element = etree.SubElement(hintgroup_element, "hintpart")
                    hintpart_element.set("on", str(hint_name))

                    hint_text_element = etree.SubElement(hintpart_element, "text")
                    hint_text_element.text = str(hint_text)

            if hint_fn:
                assert not hint_list
                hintgroup_element.set("hintfn", hint_fn)

        for additional_answer in additional_answers:
            additional_node = etree.SubElement(response_element, "additional_answer")
            additional_node.set("answer", additional_answer)

        for answer in non_attribute_answers:
            additional_node = etree.SubElement(response_element, "additional_answer")
            additional_node.text = answer

        return response_element

    def create_input_element(self, **kwargs):
        return ResponseXMLFactory.textline_input_xml(**kwargs)


class AnnotationResponseXMLFactory(ResponseXMLFactory):
    """Factory for creating <annotationresponse> XML trees"""

    def create_response_element(self, **kwargs):
        """Create a <annotationresponse> element"""
        return etree.Element("annotationresponse")

    def create_input_element(self, **kwargs):
        """Create a <annotationinput> element."""

        input_element = etree.Element("annotationinput")

        text_children = [
            {"tag": "title", "text": kwargs.get("title", "super cool annotation")},
            {"tag": "text", "text": kwargs.get("text", "texty text")},
            {"tag": "comment", "text": kwargs.get("comment", "blah blah erudite comment blah blah")},
            {"tag": "comment_prompt", "text": kwargs.get("comment_prompt", "type a commentary below")},
            {"tag": "tag_prompt", "text": kwargs.get("tag_prompt", "select one tag")},
        ]

        for child in text_children:
            etree.SubElement(input_element, child["tag"]).text = child["text"]

        default_options = [("green", "correct"), ("eggs", "incorrect"), ("ham", "partially-correct")]
        options = kwargs.get("options", default_options)
        options_element = etree.SubElement(input_element, "options")

        for description, correctness in options:
            option_element = etree.SubElement(options_element, "option", {"choice": correctness})
            option_element.text = description

        return input_element


class SymbolicResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <symbolicresponse> xml"""

    def create_response_element(self, **kwargs):
        """Build the <symbolicresponse> XML element.

        Uses ``**kwargs``:

        *expect*: The correct answer (a sympy string)

        *options*: list of option strings to pass to symmath_check
            (e.g. 'matrix', 'qbit', 'imaginary', 'numerical')"""

        # Retrieve **kwargs
        expect = kwargs.get("expect", "")
        options = kwargs.get("options", [])

        # Symmath check expects a string of options
        options_str = ",".join(options)

        # Construct the <symbolicresponse> element
        response_element = etree.Element("symbolicresponse")

        if expect:
            response_element.set("expect", str(expect))

        if options_str:
            response_element.set("options", str(options_str))

        return response_element

    def create_input_element(self, **kwargs):
        return ResponseXMLFactory.textline_input_xml(**kwargs)


class ChoiceTextResponseXMLFactory(ResponseXMLFactory):
    """Factory for producing <choicetextresponse> xml"""

    def create_response_element(self, **kwargs):
        """Create a <choicetextresponse> element"""
        return etree.Element("choicetextresponse")

    def create_input_element(self, **kwargs):
        """Create a <checkboxgroup> element.
        choices can be specified in the following format:
        [("true", [{"answer": "5", "tolerance": 0}]),
        ("false", [{"answer": "5", "tolerance": 0}])
        ]

        This indicates that the first checkbox/radio is correct and it
        contains a numtolerance_input with an answer of 5 and a tolerance of 0

        It also indicates that the second has a second incorrect radiobutton
        or checkbox with a numtolerance_input.
        """
        choices = kwargs.get("choices", [("true", {})])
        choice_inputs = []
        # Ensure that the first element of choices is an ordered
        # collection. It will start as a list, a tuple, or not a Container.
        if not isinstance(choices[0], (list, tuple)):
            choices = [choices]

        for choice in choices:
            correctness, answers = choice
            numtolerance_inputs = []
            # If the current `choice` contains any("answer": number)
            # elements, turn those into numtolerance_inputs
            if answers:
                # `answers` will be a list or tuple of answers or a single
                # answer, representing the answers for numtolerance_inputs
                # inside of this specific choice.

                # Make sure that `answers` is an ordered collection for
                # convenience.
                if not isinstance(answers, (list, tuple)):
                    answers = [answers]

                numtolerance_inputs = [self._create_numtolerance_input_element(answer) for answer in answers]

            choice_inputs.append(self._create_choice_element(correctness=correctness, inputs=numtolerance_inputs))
        # Default type is 'radiotextgroup'
        input_type = kwargs.get("type", "radiotextgroup")
        input_element = etree.Element(input_type)

        for ind, choice in enumerate(choice_inputs):
            # Give each choice text equal to it's position(0,1,2...)
            choice.text = f"choice_{ind}"
            input_element.append(choice)

        return input_element

    def _create_choice_element(self, **kwargs):
        """
        Creates a choice element for a choictextproblem.
        Defaults to a correct choice with no numtolerance_input
        """
        text = kwargs.get("text", "")
        correct = kwargs.get("correctness", "true")
        inputs = kwargs.get("inputs", [])
        choice_element = etree.Element("choice")
        choice_element.set("correct", correct)
        choice_element.text = text
        for inp in inputs:
            # Add all of the inputs as children of this choice
            choice_element.append(inp)

        return choice_element

    def _create_numtolerance_input_element(self, params):
        """
        Creates a <numtolerance_input/>  or <decoy_input/> element with
        optionally specified tolerance and answer.
        """
        answer = params["answer"] if "answer" in params else None
        # If there is not an answer specified, Then create a <decoy_input/>
        # otherwise create a <numtolerance_input/> and set its tolerance
        # and answer attributes.
        if answer:
            text_input = etree.Element("numtolerance_input")
            text_input.set("answer", answer)
            # If tolerance was specified, was specified use it, otherwise
            # Set the tolerance to "0"
            text_input.set("tolerance", params["tolerance"] if "tolerance" in params else "0")

        else:
            text_input = etree.Element("decoy_input")

        return text_input
